"""
Support for reboot, shutdown, etc on POSIX-like systems.

.. note::

    If a wrapper such as ``molly-guard`` to intercept *interactive* shutdown
    commands is configured, calling :mod:`system.halt <salt.modules.system.halt>`,
    :mod:`system.poweroff <salt.modules.system.poweroff>`,
    :mod:`system.reboot <salt.modules.system.reboot>`, and
    :mod:`system.shutdown <salt.modules.system.shutdown>` with ``salt-call`` will
    hang indefinitely while the wrapper script waits for user input. Calling them
    with ``salt`` will work as expected.
"""

import logging
import os.path
import re
from datetime import datetime, timedelta, tzinfo

import salt.utils.files
import salt.utils.path
import salt.utils.platform
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.utils.decorators import depends

log = logging.getLogger(__name__)

__virtualname__ = "system"


def __virtual__():
    """
    Only supported on POSIX-like systems
    Windows, Solaris, and OS X have their own modules
    """
    if salt.utils.platform.is_windows():
        return (False, "This module is not available on Windows")

    if salt.utils.platform.is_darwin():
        return (False, "This module is not available on Mac OS")

    if salt.utils.platform.is_sunos():
        return (False, "This module is not available on SunOS")

    return __virtualname__


def halt():
    """
    Halt a running system

    CLI Example:

    .. code-block:: bash

        salt '*' system.halt
    """
    cmd = ["halt"]
    ret = __salt__["cmd.run"](cmd, python_shell=False)
    return ret


def init(runlevel):
    """
    Change the system runlevel on sysV compatible systems

    CLI Example:

    .. code-block:: bash

        salt '*' system.init 3
    """
    cmd = ["init", f"{runlevel}"]
    ret = __salt__["cmd.run"](cmd, python_shell=False)
    return ret


def poweroff():
    """
    Poweroff a running system

    CLI Example:

    .. code-block:: bash

        salt '*' system.poweroff
    """
    cmd = ["poweroff"]
    ret = __salt__["cmd.run"](cmd, python_shell=False)
    return ret


def reboot(at_time=None):
    """
    Reboot the system

    at_time
        The wait time in minutes before the system will be rebooted.

    CLI Example:

    .. code-block:: bash

        salt '*' system.reboot
    """
    cmd = ["shutdown", "-r", (f"{at_time}" if at_time else "now")]
    ret = __salt__["cmd.run"](cmd, python_shell=False)
    return ret


def shutdown(at_time=None):
    """
    Shutdown a running system

    at_time
        The wait time in minutes before the system will be shutdown.

    CLI Example:

    .. code-block:: bash

        salt '*' system.shutdown 5
    """
    if (
        salt.utils.platform.is_freebsd()
        or salt.utils.platform.is_netbsd()
        or salt.utils.platform.is_openbsd()
    ):
        # these platforms don't power off by default when halted
        flag = "-p"
    else:
        flag = "-h"

    cmd = ["shutdown", flag, (f"{at_time}" if at_time else "now")]
    ret = __salt__["cmd.run"](cmd, python_shell=False)
    return ret


def _date_bin_set_datetime(new_date):
    """
    set the system date/time using the date command

    Note using a strictly POSIX-compliant date binary we can only set the date
    up to the minute.
    """
    cmd = ["date"]

    # if there is a timezone in the datetime object use that offset
    # This will modify the new_date to be the equivalent time in UTC
    if new_date.utcoffset() is not None:
        new_date = new_date - new_date.utcoffset()
        new_date = new_date.replace(tzinfo=_FixedOffset(0))
        cmd.append("-u")

    # the date can be set in the following format:
    # Note that setting the time with a resolution of seconds
    # is not a posix feature, so we will attempt it and if it
    # fails we will try again only using posix features

    # date MMDDhhmm[[CC]YY[.ss]]
    non_posix = "{1:02}{2:02}{3:02}{4:02}{0:04}.{5:02}".format(*new_date.timetuple())
    non_posix_cmd = cmd + [non_posix]

    ret_non_posix = __salt__["cmd.run_all"](non_posix_cmd, python_shell=False)
    if ret_non_posix["retcode"] != 0:
        # We will now try the command again following posix
        # date MMDDhhmm[[CC]YY]
        posix = " {1:02}{2:02}{3:02}{4:02}{0:04}".format(*new_date.timetuple())
        posix_cmd = cmd + [posix]

        ret_posix = __salt__["cmd.run_all"](posix_cmd, python_shell=False)
        if ret_posix["retcode"] != 0:
            # if both fail it's likely an invalid date string
            # so we will give back the error from the first attempt
            msg = "date failed: {}".format(ret_non_posix["stderr"])
            raise CommandExecutionError(msg)
    return True


def has_settable_hwclock():
    """
    Returns ``True`` if the system has a hardware clock capable of being
    set from software.

    CLI Example:

    .. code-block:: bash

        salt '*' system.has_settable_hwclock
    """
    if salt.utils.path.which_bin(["hwclock"]) is not None:
        res = __salt__["cmd.run_all"](
            ["hwclock", "--test", "--systohc"],
            python_shell=False,
            output_loglevel="quiet",
            ignore_retcode=True,
        )
        return res["retcode"] == 0
    return False


def _swclock_to_hwclock():
    """
    Set hardware clock to value of software clock.
    """
    res = __salt__["cmd.run_all"](["hwclock", "--systohc"], python_shell=False)
    if res["retcode"] != 0:
        log.warn(
            "hwclock failed to set hardware clock from software clock: %s",
            res["stderr"],
        )
    return True


def _try_parse_datetime(time_str, fmts):
    """
    Attempts to parse the input ``time_str`` as a date.

    :param str time_str: A string representing the time
    :param list fmts: A list of date format strings.

    :return: Returns a datetime object if parsed properly. Otherwise ``None``
    :rtype: datetime
    """
    result = None
    for fmt in fmts:
        try:
            result = datetime.strptime(time_str, fmt)
            break
        except ValueError:
            pass
    return result


def _offset_to_min(utc_offset):
    """
    Helper function that converts the UTC offset string into number of minutes
    offset. Input is in form ``[+-]?HHMM``. Example valid inputs are ``+0500``
    ``-0300`` and ``0800``. These would return ``-300``, ``180``, ``480`` respectively.
    """
    match = re.match(r"^([+-])?(\d\d)(\d\d)$", utc_offset)
    if not match:
        raise SaltInvocationError("Invalid UTC offset")

    sign = -1 if match.group(1) == "-" else 1
    hours_offset = int(match.group(2))
    minutes_offset = int(match.group(3))
    total_offset = sign * (hours_offset * 60 + minutes_offset)
    return total_offset


def _get_offset_time(utc_offset):
    """
    Will return the current time adjusted using the input timezone offset.

    :rtype: datetime
    """
    if utc_offset is not None:
        minutes = _offset_to_min(utc_offset)
        offset = timedelta(minutes=minutes)
        offset_time = datetime.utcnow() + offset
        offset_time = offset_time.replace(tzinfo=_FixedOffset(minutes))
    else:
        offset_time = datetime.now()
    return offset_time


def get_system_time(utc_offset=None):
    """
    Get the system time.

    :param str utc_offset: The UTC offset in 4 digit (e.g. ``+0600``) format with an
        optional sign (``+``/``-``).  Will default to ``None`` which will use the local
        timezone. To set the time based off of UTC use ``+0000``. Note: If
        being passed through the command line will need to be quoted twice to
        allow negative offsets (e.g. ``"'+0000'"``).
    :return: Returns the system time in ``HH:MM:SS AM/PM`` format.
    :rtype: str

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_system_time
    """
    offset_time = _get_offset_time(utc_offset)
    return datetime.strftime(offset_time, "%I:%M:%S %p")


def set_system_time(newtime, utc_offset=None):
    """
    Set the system time.

    :param str newtime:
        The time to set. Can be any of the following formats.

        - ``HH:MM:SS AM/PM``
        - ``HH:MM AM/PM``
        - ``HH:MM:SS`` (24 hour)
        - ``HH:MM`` (24 hour)

        Note that the Salt command line parser parses the date/time
        before we obtain the argument (preventing us from doing UTC)
        Therefore the argument must be passed in as a string.
        Meaning the text might have to be quoted twice on the command line.

    :param str utc_offset: The UTC offset in 4 digit (``+0600``) format with an
        optional sign (``+``/``-``).  Will default to ``None`` which will use the local
        timezone. To set the time based off of UTC use ``+0000``. Note: If
        being passed through the command line will need to be quoted twice to
        allow negative offsets (e.g. ``"'+0000'"``)
    :return: Returns ``True`` if successful. Otherwise ``False``.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' system.set_system_time "'11:20'"
    """
    fmts = ["%I:%M:%S %p", "%I:%M %p", "%H:%M:%S", "%H:%M"]
    dt_obj = _try_parse_datetime(newtime, fmts)
    if dt_obj is None:
        return False

    return set_system_date_time(
        hours=dt_obj.hour,
        minutes=dt_obj.minute,
        seconds=dt_obj.second,
        utc_offset=utc_offset,
    )


def get_system_date_time(utc_offset=None):
    """
    Get the system date/time.

    :param str utc_offset: The UTC offset in 4 digit (``+0600``) format with an
        optional sign (``+``/``-``).  Will default to ``None`` which will use the local
        timezone. To set the time based off of UTC use ``+0000``. Note: If
        being passed through the command line will need to be quoted twice to
        allow negative offsets (e.g. ``"'+0000'"``).
    :return: Returns the system time in ``YYYY-MM-DD hh:mm:ss`` format.
    :rtype: str

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_system_date_time "'-0500'"
    """
    offset_time = _get_offset_time(utc_offset)
    return datetime.strftime(offset_time, "%Y-%m-%d %H:%M:%S")


def set_system_date_time(
    years=None,
    months=None,
    days=None,
    hours=None,
    minutes=None,
    seconds=None,
    utc_offset=None,
):
    """
    Set the system date and time. Each argument is an element of the date, but
    not required. If an element is not passed, the current system value for
    that element will be used. For example, if the year is not passed, the
    current system year will be used. (Used by
    :mod:`system.set_system_date <salt.modules.system.set_system_date>` and
    :mod:`system.set_system_time <salt.modules.system.set_system_time>`)

    Updates hardware clock, if present, in addition to software
    (kernel) clock.

    :param int years: Years digit, e.g.: ``2015``
    :param int months: Months digit: ``1``-``12``
    :param int days: Days digit: ``1``-``31``
    :param int hours: Hours digit: ``0``-``23``
    :param int minutes: Minutes digit: ``0``-``59``
    :param int seconds: Seconds digit: ``0``-``59``
    :param str utc_offset: The UTC offset in 4 digit (``+0600``) format with an
        optional sign (``+``/``-``).  Will default to ``None`` which will use the local
        timezone. To set the time based off of UTC use ``+0000``. Note: If
        being passed through the command line will need to be quoted twice to
        allow negative offsets (e.g. ``"'+0000'"``).
    :return: ``True`` if successful. Otherwise ``False``.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' system.set_system_date_time 2015 5 12 11 37 53 "'-0500'"
    """
    # Get the current date/time
    date_time = _get_offset_time(utc_offset)

    # Check for passed values. If not passed, use current values
    if years is None:
        years = date_time.year
    if months is None:
        months = date_time.month
    if days is None:
        days = date_time.day
    if hours is None:
        hours = date_time.hour
    if minutes is None:
        minutes = date_time.minute
    if seconds is None:
        seconds = date_time.second

    try:
        new_datetime = datetime(
            years, months, days, hours, minutes, seconds, 0, date_time.tzinfo
        )
    except ValueError as err:
        raise SaltInvocationError(err.message)

    if not _date_bin_set_datetime(new_datetime):
        return False

    if has_settable_hwclock():
        # Now that we've successfully set the software clock, we should
        # update hardware clock for time to persist though reboot.
        return _swclock_to_hwclock()

    return True


def get_system_date(utc_offset=None):
    """
    Get the system date

    :param str utc_offset: The UTC offset in 4 digit (``+0600``) format with an
        optional sign (``+``/``-``).  Will default to ``None`` which will use the local
        timezone. To set the time based off of UTC use ``+0000``. Note: If
        being passed through the command line will need to be quoted twice to
        allow negative offsets (e.g. ``"'+0000'"``).
    :return: Returns the system date.
    :rtype: str

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_system_date
    """
    offset_time = _get_offset_time(utc_offset)
    return datetime.strftime(offset_time, "%a %m/%d/%Y")


def set_system_date(newdate, utc_offset=None):
    """
    Set the system date. Use ``<mm-dd-yy>`` format for the date.

    :param str newdate:
        The date to set. Can be any of the following formats:

        - ``YYYY-MM-DD``
        - ``MM-DD-YYYY``
        - ``MM-DD-YY``
        - ``MM/DD/YYYY``
        - ``MM/DD/YY``
        - ``YYYY/MM/DD``

    CLI Example:

    .. code-block:: bash

        salt '*' system.set_system_date '03-28-13'
    """
    fmts = ["%Y-%m-%d", "%m-%d-%Y", "%m-%d-%y", "%m/%d/%Y", "%m/%d/%y", "%Y/%m/%d"]

    # Get date/time object from newdate
    dt_obj = _try_parse_datetime(newdate, fmts)
    if dt_obj is None:
        raise SaltInvocationError("Invalid date format")

    # Set time using set_system_date_time()
    return set_system_date_time(
        years=dt_obj.year, months=dt_obj.month, days=dt_obj.day, utc_offset=utc_offset
    )


# Class from: <https://docs.python.org/2.7/library/datetime.html>

# A class building tzinfo objects for fixed-offset time zones.
# Note that _FixedOffset(0) is a way to build a UTC tzinfo object.


class _FixedOffset(tzinfo):
    """
    Fixed offset in minutes east from UTC.
    """

    def __init__(self, offset):
        super().__init__()
        self.__offset = timedelta(minutes=offset)

    def utcoffset(self, dt):  # pylint: disable=W0613
        return self.__offset

    def tzname(self, dt):  # pylint: disable=W0613
        return None

    def dst(self, dt):  # pylint: disable=W0613
        return timedelta(0)


def _strip_quotes(str_q):
    """
    Helper function to strip off the ``'`` or ``"`` off of a string
    """
    if str_q[0] == str_q[-1] and str_q.startswith(("'", '"')):
        return str_q[1:-1]
    return str_q


def get_computer_desc():
    """
    Get ``PRETTY_HOSTNAME`` value stored in ``/etc/machine-info``
    If this file doesn't exist or the variable doesn't exist
    return ``False``.

    :return: Value of ``PRETTY_HOSTNAME`` in ``/etc/machine-info``.
        If file/variable does not exist ``False``.
    :rtype: str

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_computer_desc
    """
    hostname_cmd = salt.utils.path.which("hostnamectl")
    if hostname_cmd:
        desc = __salt__["cmd.run"](
            [hostname_cmd, "status", "--pretty"], python_shell=False
        )
    else:
        desc = None
        pattern = re.compile(r"^\s*PRETTY_HOSTNAME=(.*)$")
        try:
            with salt.utils.files.fopen("/etc/machine-info", "r") as mach_info:
                for line in mach_info.readlines():
                    line = salt.utils.stringutils.to_unicode(line)
                    match = pattern.match(line)
                    if match:
                        # get rid of whitespace then strip off quotes
                        desc = _strip_quotes(match.group(1).strip())
                        # no break so we get the last occurance
        except OSError:
            pass

        if desc is None:
            return False

    return (
        desc.replace(r"\\", "\\")
        .replace(r"\"", r'"')
        .replace(r"\n", "\n")
        .replace(r"\t", "\t")
    )


def set_computer_desc(desc):
    """
    Set ``PRETTY_HOSTNAME`` value stored in ``/etc/machine-info``
    This will create the file if it does not exist. If
    it is unable to create or modify this file, ``False`` is returned.

    :param str desc: The computer description
    :return: ``False`` on failure. ``True`` if successful.

    CLI Example:

    .. code-block:: bash

        salt '*' system.set_computer_desc "Michael's laptop"
    """
    desc = (
        salt.utils.stringutils.to_unicode(desc)
        .replace('"', r"\"")
        .replace("\n", r"\n")
        .replace("\t", r"\t")
    )

    hostname_cmd = salt.utils.path.which("hostnamectl")
    if hostname_cmd:
        result = __salt__["cmd.retcode"](
            [hostname_cmd, "set-hostname", "--pretty", desc], python_shell=False
        )
        return True if result == 0 else False

    if not os.path.isfile("/etc/machine-info"):
        with salt.utils.files.fopen("/etc/machine-info", "w"):
            pass

    pattern = re.compile(r"^\s*PRETTY_HOSTNAME=(.*)$")
    new_line = salt.utils.stringutils.to_str(f'PRETTY_HOSTNAME="{desc}"')
    try:
        with salt.utils.files.fopen("/etc/machine-info", "r+") as mach_info:
            lines = mach_info.readlines()
            for i, line in enumerate(lines):
                if pattern.match(salt.utils.stringutils.to_unicode(line)):
                    lines[i] = new_line
                    break
            else:
                # PRETTY_HOSTNAME line was not found, add it to the end
                lines.append(new_line)
            # time to write our changes to the file
            mach_info.seek(0, 0)
            mach_info.truncate()
            mach_info.writelines(lines)
            return True
    except OSError:
        return False


def set_computer_name(hostname):
    """
    Modify hostname.

    CLI Example:

    .. code-block:: bash

        salt '*' system.set_computer_name master.saltstack.com
    """
    return __salt__["network.mod_hostname"](hostname)


def get_computer_name():
    """
    Get hostname.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_computer_name
    """
    return __salt__["network.get_hostname"]()


def _is_nilrt_family():
    """
    Determine whether the minion is running on NI Linux RT
    """
    return __grains__.get("os_family") == "NILinuxRT"


NILRT_REBOOT_WITNESS_PATH = "/var/volatile/tmp/salt/reboot_witnessed"


@depends("_is_nilrt_family")
def set_reboot_required_witnessed():
    """
    .. note::

        This only applies to Minions running on NI Linux RT

    This function is used to remember that an event indicating that a reboot is
    required was witnessed. This function writes to a temporary filesystem so
    the event gets cleared upon reboot.

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' system.set_reboot_required_witnessed
    """
    errcode = -1
    dir_path = os.path.dirname(NILRT_REBOOT_WITNESS_PATH)
    if not os.path.exists(dir_path):
        try:
            os.makedirs(dir_path)
        except OSError as ex:
            raise SaltInvocationError(
                f"Error creating {dir_path} (-{ex.errno}): {ex.strerror}"
            )

        rdict = __salt__["cmd.run_all"](f"touch {NILRT_REBOOT_WITNESS_PATH}")
        errcode = rdict["retcode"]

    return errcode == 0


@depends("_is_nilrt_family")
def get_reboot_required_witnessed():
    """
    .. note::

        This only applies to Minions running on NI Linux RT

    Determine if at any time during the current boot session the salt minion
    witnessed an event indicating that a reboot is required.

    Returns:
        bool: ``True`` if the a reboot request was witnessed, ``False`` otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_reboot_required_witnessed
    """
    return os.path.exists(NILRT_REBOOT_WITNESS_PATH)



"""
Swift utility class
===================
Author: Anthony Stanton <anthony.stanton@gmail.com>
"""

import logging
import sys
from errno import EEXIST
from os import makedirs
from os.path import dirname, isdir

import salt.utils.files

# Get logging started
log = logging.getLogger(__name__)

HAS_SWIFT = False
try:
    from swiftclient import client

    HAS_SWIFT = True
except ImportError:
    pass


def check_swift():
    return HAS_SWIFT


def mkdirs(path):
    try:
        makedirs(path)
    except OSError as err:
        if err.errno != EEXIST:
            raise


# we've been playing fast and loose with kwargs, but the swiftclient isn't
# going to accept any old thing
def _sanitize(kwargs):
    variables = (
        "user",
        "key",
        "authurl",
        "retries",
        "preauthurl",
        "preauthtoken",
        "snet",
        "starting_backoff",
        "max_backoff",
        "tenant_name",
        "os_options",
        "auth_version",
        "cacert",
        "insecure",
        "ssl_compression",
    )
    ret = {}
    for var in kwargs:
        if var in variables:
            ret[var] = kwargs[var]

    return ret


class SaltSwift:
    """
    Class for all swiftclient functions
    """

    def __init__(
        self, user, tenant_name, auth_url, password=None, auth_version=2, **kwargs
    ):
        """
        Set up openstack credentials
        """
        if not HAS_SWIFT:
            log.error(
                "Error:: unable to find swiftclient. Try installing it from the"
                " appropriate repository."
            )
            return None

        self.kwargs = kwargs.copy()
        self.kwargs["user"] = user
        self.kwargs["password"] = password
        self.kwargs["tenant_name"] = tenant_name
        self.kwargs["authurl"] = auth_url
        self.kwargs["auth_version"] = auth_version
        if "key" not in self.kwargs:
            self.kwargs["key"] = password

        self.kwargs = _sanitize(self.kwargs)

        self.conn = client.Connection(**self.kwargs)

    def get_account(self):
        """
        List Swift containers
        """
        try:
            listing = self.conn.get_account()
            return listing
        except Exception as exc:  # pylint: disable=broad-except
            log.error("There was an error::")
            if hasattr(exc, "code") and hasattr(exc, "msg"):
                log.error("    Code: %s: %s", exc.code, exc.msg)
            log.error("    Content: \n%s", getattr(exc, "read", lambda: str(exc))())
            return False

    def get_container(self, cont):
        """
        List files in a Swift container
        """
        try:
            listing = self.conn.get_container(cont)
            return listing
        except Exception as exc:  # pylint: disable=broad-except
            log.error("There was an error::")
            if hasattr(exc, "code") and hasattr(exc, "msg"):
                log.error("    Code: %s: %s", exc.code, exc.msg)
            log.error("    Content: \n%s", getattr(exc, "read", lambda: str(exc))())
            return False

    def put_container(self, cont):
        """
        Create a new Swift container
        """
        try:
            self.conn.put_container(cont)
            return True
        except Exception as exc:  # pylint: disable=broad-except
            log.error("There was an error::")
            if hasattr(exc, "code") and hasattr(exc, "msg"):
                log.error("    Code: %s: %s", exc.code, exc.msg)
            log.error("    Content: \n%s", getattr(exc, "read", lambda: str(exc))())
            return False

    def delete_container(self, cont):
        """
        Delete a Swift container
        """
        try:
            self.conn.delete_container(cont)
            return True
        except Exception as exc:  # pylint: disable=broad-except
            log.error("There was an error::")
            if hasattr(exc, "code") and hasattr(exc, "msg"):
                log.error("    Code: %s: %s", exc.code, exc.msg)
            log.error("    Content: \n%s", getattr(exc, "read", lambda: str(exc))())
            return False

    def post_container(self, cont, metadata=None):
        """
        Update container metadata
        """

    def head_container(self, cont):
        """
        Get container metadata
        """

    def get_object(self, cont, obj, local_file=None, return_bin=False):
        """
        Retrieve a file from Swift
        """
        try:
            if local_file is None and return_bin is False:
                return False

            headers, body = self.conn.get_object(cont, obj, resp_chunk_size=65536)

            if return_bin is True:
                fp = sys.stdout
            else:
                dirpath = dirname(local_file)
                if dirpath and not isdir(dirpath):
                    mkdirs(dirpath)
                # pylint: disable=resource-leakage
                fp = salt.utils.files.fopen(local_file, "wb")
                # pylint: enable=resource-leakage

            read_length = 0
            for chunk in body:
                read_length += len(chunk)
                fp.write(chunk)
            fp.close()
            return True

        # ClientException
        # file/dir exceptions
        except Exception as exc:  # pylint: disable=broad-except
            log.error("There was an error::")
            if hasattr(exc, "code") and hasattr(exc, "msg"):
                log.error("    Code: %s: %s", exc.code, exc.msg)
            log.error("    Content: \n%s", getattr(exc, "read", lambda: str(exc))())
            return False

    def put_object(self, cont, obj, local_file):
        """
        Upload a file to Swift
        """
        try:
            with salt.utils.files.fopen(local_file, "rb") as fp_:
                self.conn.put_object(cont, obj, fp_)
            return True
        except Exception as exc:  # pylint: disable=broad-except
            log.error("There was an error::")
            if hasattr(exc, "code") and hasattr(exc, "msg"):
                log.error("    Code: %s: %s", exc.code, exc.msg)
            log.error("    Content: \n%s", getattr(exc, "read", lambda: str(exc))())
            return False

    def delete_object(self, cont, obj):
        """
        Delete a file from Swift
        """
        try:
            self.conn.delete_object(cont, obj)
            return True
        except Exception as exc:  # pylint: disable=broad-except
            log.error("There was an error::")
            if hasattr(exc, "code") and hasattr(exc, "msg"):
                log.error("    Code: %s: %s", exc.code, exc.msg)
            log.error("    Content: \n%s", getattr(exc, "read", lambda: str(exc))())
            return False

    def head_object(self, cont, obj):
        """
        Get object metadata
        """

    def post_object(self, cont, obj, metadata):
        """
        Update object metadata
        """



"""
    Sentry Logging Handler
    ======================

    .. versionadded:: 0.17.0

    This module provides a `Sentry`_ logging handler. Sentry is an open source
    error tracking platform that provides deep context about exceptions that
    happen in production. Details about stack traces along with the context
    variables available at the time of the exception are easily browsable and
    filterable from the online interface. For more details please see
    `Sentry`_.

    .. admonition:: Note

        The `Raven`_ library needs to be installed on the system for this
        logging handler to be available.

    Configuring the python `Sentry`_ client, `Raven`_, should be done under the
    ``sentry_handler`` configuration key. Additional `context` may be provided
    for corresponding grain item(s).
    At the bare minimum, you need to define the `DSN`_. As an example:

    .. code-block:: yaml

        sentry_handler:
          dsn: https://pub-key:secret-key@app.getsentry.com/app-id


    More complex configurations can be achieved, for example:

    .. code-block:: yaml

        sentry_handler:
          servers:
            - https://sentry.example.com
            - http://192.168.1.1
          project: app-id
          public_key: deadbeefdeadbeefdeadbeefdeadbeef
          secret_key: beefdeadbeefdeadbeefdeadbeefdead
          context:
            - os
            - master
            - saltversion
            - cpuarch
            - ec2.tags.environment

    .. admonition:: Note

        The ``public_key`` and ``secret_key`` variables are not supported with
        Sentry > 3.0. The `DSN`_ key should be used instead.

    All the client configuration keys are supported, please see the
    `Raven client documentation`_.

    The default logging level for the sentry handler is ``ERROR``. If you wish
    to define a different one, define ``log_level`` under the
    ``sentry_handler`` configuration key:

    .. code-block:: yaml

      sentry_handler:
        dsn: https://pub-key:secret-key@app.getsentry.com/app-id
        log_level: warning


    The available log levels are those also available for the salt ``cli``
    tools and configuration; ``salt --help`` should give you the required
    information.


    Threaded Transports
    -------------------

    Raven's documents rightly suggest using its threaded transport for
    critical applications. However, don't forget that if you start having
    troubles with Salt after enabling the threaded transport, please try
    switching to a non-threaded transport to see if that fixes your problem.



    .. _`DSN`: https://raven.readthedocs.io/en/latest/config/index.html#the-sentry-dsn
    .. _`Sentry`: https://getsentry.com
    .. _`Raven`: https://raven.readthedocs.io
    .. _`Raven client documentation`: https://raven.readthedocs.io/en/latest/config/index.html#client-arguments
"""

import logging
import re

import salt.loader
from salt._logging import LOG_LEVELS

try:
    import raven
    from raven.handlers.logging import SentryHandler

    HAS_RAVEN = True
except ImportError:
    HAS_RAVEN = False

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "sentry"


def __virtual__():
    load_err_msg = []
    if not HAS_RAVEN:
        load_err_msg.append("Cannot find 'raven' python library")
    if not __opts__.get("sentry_handler"):
        load_err_msg.append("'sentry_handler' config is empty or not defined")
    if load_err_msg:
        return False, ", ".join(load_err_msg)
    return __virtualname__


def setup_handlers():
    """
    sets up the sentry handler
    """
    if not __opts__.get("sentry_handler"):
        log.debug("'sentry_handler' config is empty or not defined")
        return False

    # Regenerating dunders can be expensive, so only do it if the user enables
    # `sentry_handler` as checked above
    __grains__ = salt.loader.grains(__opts__)
    __salt__ = salt.loader.minion_mods(__opts__)

    options = {}
    dsn = get_config_value("dsn")
    if dsn is not None:
        try:
            # support raven ver 5.5.0
            from raven.transport import TransportRegistry, default_transports
            from raven.utils.urlparse import urlparse

            transport_registry = TransportRegistry(default_transports)
            url = urlparse(dsn)
            if not transport_registry.supported_scheme(url.scheme):
                raise ValueError(f"Unsupported Sentry DSN scheme: {url.scheme}")
        except ValueError as exc:
            log.info("Raven failed to parse the configuration provided DSN: %s", exc)

    if not dsn:
        for key in ("project", "servers", "public_key", "secret_key"):
            config_value = get_config_value(key)
            if config_value is None and key not in options:
                log.debug(
                    "The required 'sentry_handler' configuration key, "
                    "'%s', is not properly configured. Not configuring "
                    "the sentry logging handler.",
                    key,
                )
                return
            elif config_value is None:
                continue
            options[key] = config_value

    # site: An optional, arbitrary string to identify this client installation.
    options.update(
        {
            # site: An optional, arbitrary string to identify this client
            # installation
            "site": get_config_value("site"),
            # name: This will override the server_name value for this installation.
            # Defaults to socket.gethostname()
            "name": get_config_value("name"),
            # exclude_paths: Extending this allow you to ignore module prefixes
            # when sentry attempts to discover which function an error comes from
            "exclude_paths": get_config_value("exclude_paths", ()),
            # include_paths: For example, in Django this defaults to your list of
            # INSTALLED_APPS, and is used for drilling down where an exception is
            # located
            "include_paths": get_config_value("include_paths", ()),
            # list_max_length: The maximum number of items a list-like container
            # should store.
            "list_max_length": get_config_value("list_max_length"),
            # string_max_length: The maximum characters of a string that should be
            # stored.
            "string_max_length": get_config_value("string_max_length"),
            # auto_log_stacks: Should Raven automatically log frame stacks
            # (including locals) all calls as it would for exceptions.
            "auto_log_stacks": get_config_value("auto_log_stacks"),
            # timeout: If supported, the timeout value for sending messages to
            # remote.
            "timeout": get_config_value("timeout", 1),
            # processors: A list of processors to apply to events before sending
            # them to the Sentry server. Useful for sending additional global state
            # data or sanitizing data that you want to keep off of the server.
            "processors": get_config_value("processors"),
            # dsn: Ensure the DSN is passed into the client
            "dsn": dsn,
        }
    )

    client = raven.Client(**options)
    context = get_config_value("context")
    context_dict = {}
    if context is not None:
        for tag in context:
            try:
                tag_value = __grains__[tag]
            except KeyError:
                log.debug("Sentry tag '%s' not found in grains.", tag)
                continue
            if tag_value:
                context_dict[tag] = tag_value
        if context_dict:
            client.context.merge({"tags": context_dict})
    try:
        handler = SentryHandler(client)

        exclude_patterns = get_config_value("exclude_patterns", None)
        if exclude_patterns:
            filter_regexes = [re.compile(pattern) for pattern in exclude_patterns]

            class FilterExcludedMessages:
                @staticmethod
                def filter(record):
                    m = record.getMessage()
                    return not any(regex.search(m) for regex in filter_regexes)

            handler.addFilter(FilterExcludedMessages())

        handler.setLevel(LOG_LEVELS[get_config_value("log_level", "error")])
        return handler
    except ValueError as exc:
        log.debug("Failed to setup the sentry logging handler", exc_info=True)


def get_config_value(name, default=None):
    """
    returns a configuration option for the sentry_handler
    """
    return __opts__["sentry_handler"].get(name, default)



import logging
import random
import string

import pytest

import salt.config
import salt.loader
import salt.modules.boto_cognitoidentity as boto_cognitoidentity
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

# pylint: disable=import-error,no-name-in-module
try:
    import boto3
    from botocore.exceptions import ClientError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
]


# pylint: enable=import-error,no-name-in-module

# the boto_lambda module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto3_version = "1.2.1"

region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}
error_message = (
    "An error occurred (101) when calling the {0} operation: Test-defined error"
)
error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}

first_pool_id = "first_pool_id"
first_pool_name = "first_pool"
second_pool_id = "second_pool_id"
second_pool_name = "second_pool"
second_pool_name_updated = "second_pool_updated"
third_pool_id = "third_pool_id"
third_pool_name = first_pool_name

identity_pools_ret = dict(
    IdentityPools=[
        dict(IdentityPoolId=first_pool_id, IdentityPoolName=first_pool_name),
        dict(IdentityPoolId=second_pool_id, IdentityPoolName=second_pool_name),
        dict(IdentityPoolId=third_pool_id, IdentityPoolName=third_pool_name),
    ]
)

first_pool_ret = dict(
    IdentityPoolId=first_pool_id,
    IdentityPoolName=first_pool_name,
    AllowUnauthenticatedIdentities=False,
    SupportedLoginProviders={
        "accounts.google.com": "testing123",
        "api.twitter.com": "testing123",
        "graph.facebook.com": "testing123",
        "www.amazon.com": "testing123",
    },
    DeveloperProviderName="test_provider",
    OpenIdConnectProviderARNs=["some_provider_arn", "another_provider_arn"],
)

first_pool_role_ret = dict(
    IdentityPoolId=first_pool_id,
    Roles=dict(
        authenticated="first_pool_auth_role", unauthenticated="first_pool_unauth_role"
    ),
)

second_pool_ret = dict(
    IdentityPoolId=second_pool_id,
    IdentityPoolName=second_pool_name,
    AllowUnauthenticatedIdentities=False,
)

third_pool_ret = dict(
    IdentityPoolId=third_pool_id,
    IdentityPoolName=third_pool_name,
    AllowUnauthenticatedIdentities=False,
    DeveloperProviderName="test_provider2",
)

third_pool_role_ret = dict(IdentityPoolId=third_pool_id)

default_pool_ret = dict(
    IdentityPoolId="default_pool_id",
    IdentityPoolName="default_pool_name",
    AllowUnauthenticatedIdentities=False,
    DeveloperProviderName="test_provider_default",
)

default_pool_role_ret = dict(IdentityPoolId="default_pool_id")

log = logging.getLogger(__name__)


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif Version(boto3.__version__) < Version(required_boto3_version):
        return False
    else:
        return True


class BotoCognitoIdentityTestCaseBase(TestCase, LoaderModuleMockMixin):
    conn = None

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(
            opts, whitelist=["boto3", "args", "systemd", "path", "platform"], context={}
        )
        return {boto_cognitoidentity: {"__utils__": utils}}

    def setUp(self):
        super().setUp()
        boto_cognitoidentity.__init__(self.opts)
        del self.opts

        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        self.addCleanup(delattr, self, "conn")
        session_instance.client.return_value = self.conn


class BotoCognitoIdentityTestCaseMixin:
    pass


# @pytest.mark.skip(reason="Skip these tests while investigating failures")
@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto3 module must be greater than or equal to version {}".format(
        required_boto3_version
    ),
)
class BotoCognitoIdentityTestCase(
    BotoCognitoIdentityTestCaseBase, BotoCognitoIdentityTestCaseMixin
):
    """
    TestCase for salt.modules.boto_cognitoidentity module
    """

    def _describe_identity_pool_side_effect(self, *args, **kwargs):
        if kwargs.get("IdentityPoolId") == first_pool_id:
            return first_pool_ret
        elif kwargs.get("IdentityPoolId") == third_pool_id:
            return third_pool_ret
        else:
            return default_pool_ret

    def test_that_when_describing_a_named_identity_pool_and_pool_exists_the_describe_identity_pool_method_returns_pools_properties(
        self,
    ):
        """
        Tests describing identity pool when the pool's name exists
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        result = boto_cognitoidentity.describe_identity_pools(
            IdentityPoolName=first_pool_name, **conn_parameters
        )
        self.assertEqual(result.get("identity_pools"), [first_pool_ret, third_pool_ret])

    def test_that_when_describing_a_identity_pool_by_its_id_and_pool_exists_the_desribe_identity_pool_method_returns_pools_properties(
        self,
    ):
        """
        Tests describing identity pool when the given pool's id exists
        """
        self.conn.describe_identity_pool.return_value = third_pool_ret
        result = boto_cognitoidentity.describe_identity_pools(
            IdentityPoolName="", IdentityPoolId=third_pool_id, **conn_parameters
        )
        self.assertEqual(result.get("identity_pools"), [third_pool_ret])
        self.assertTrue(self.conn.list_identity_pools.call_count == 0)

    def test_that_when_describing_a_named_identity_pool_and_pool_does_not_exist_the_describe_identity_pool_method_returns_none(
        self,
    ):
        """
        Tests describing identity pool when the pool's name doesn't exist
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.return_value = first_pool_ret
        result = boto_cognitoidentity.describe_identity_pools(
            IdentityPoolName="no_such_pool", **conn_parameters
        )
        self.assertEqual(result.get("identity_pools", "no such key"), None)

    def test_that_when_describing_a_named_identity_pool_and_error_thrown_the_describe_identity_pool_method_returns_error(
        self,
    ):
        """
        Tests describing identity pool returns error when there is an exception to boto3 calls
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = ClientError(
            error_content, "error on describe identity pool"
        )
        result = boto_cognitoidentity.describe_identity_pools(
            IdentityPoolName=first_pool_name, **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("error on describe identity pool"),
        )

    def test_that_when_create_identity_pool_the_create_identity_pool_method_returns_created_identity_pool(
        self,
    ):
        """
        Tests the positive case where create identity pool succeeds
        """
        return_val = default_pool_ret.copy()
        return_val.pop("DeveloperProviderName", None)
        self.conn.create_identity_pool.return_value = return_val
        result = boto_cognitoidentity.create_identity_pool(
            IdentityPoolName="default_pool_name", **conn_parameters
        )
        mock_calls = self.conn.mock_calls
        # name, args, kwargs = mock_calls[0]
        self.assertTrue(result.get("created"))
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0][0], "create_identity_pool")
        self.assertNotIn("DeveloperProviderName", mock_calls[0][2])

    def test_that_when_create_identity_pool_and_error_thrown_the_create_identity_pool_method_returns_error(
        self,
    ):
        """
        Tests the negative case where create identity pool has a boto3 client error
        """
        self.conn.create_identity_pool.side_effect = ClientError(
            error_content, "create_identity_pool"
        )
        result = boto_cognitoidentity.create_identity_pool(
            IdentityPoolName="default_pool_name", **conn_parameters
        )
        self.assertIs(result.get("created"), False)
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("create_identity_pool"),
        )

    def test_that_when_delete_identity_pools_with_multiple_matching_pool_names_the_delete_identity_pools_methos_returns_true_and_deleted_count(
        self,
    ):
        """
        Tests that given 2 matching pool ids, the operation returns deleted status of true and
        count 2
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.delete_identity_pool.return_value = None
        result = boto_cognitoidentity.delete_identity_pools(
            IdentityPoolName=first_pool_name, **conn_parameters
        )
        mock_calls = self.conn.mock_calls
        self.assertTrue(result.get("deleted"))
        self.assertEqual(result.get("count"), 2)
        self.assertEqual(len(mock_calls), 3)
        self.assertEqual(mock_calls[1][0], "delete_identity_pool")
        self.assertEqual(mock_calls[2][0], "delete_identity_pool")
        self.assertEqual(mock_calls[1][2].get("IdentityPoolId"), first_pool_id)
        self.assertEqual(mock_calls[2][2].get("IdentityPoolId"), third_pool_id)

    def test_that_when_delete_identity_pools_with_no_matching_pool_names_the_delete_identity_pools_method_returns_false(
        self,
    ):
        """
        Tests that the given pool name does not exist, the operation returns deleted status of false
        and count 0
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        result = boto_cognitoidentity.delete_identity_pools(
            IdentityPoolName="no_such_pool_name", **conn_parameters
        )
        mock_calls = self.conn.mock_calls
        self.assertIs(result.get("deleted"), False)
        self.assertEqual(result.get("count"), 0)
        self.assertEqual(len(mock_calls), 1)

    def test_that_when_delete_identity_pools_and_error_thrown_the_delete_identity_pools_method_returns_false_and_the_error(
        self,
    ):
        """
        Tests that the delete_identity_pool method throws an exception
        """
        self.conn.delete_identity_pool.side_effect = ClientError(
            error_content, "delete_identity_pool"
        )
        # try to delete an unexistent pool id "no_such_pool_id"
        result = boto_cognitoidentity.delete_identity_pools(
            IdentityPoolName=first_pool_name,
            IdentityPoolId="no_such_pool_id",
            **conn_parameters
        )
        mock_calls = self.conn.mock_calls
        self.assertIs(result.get("deleted"), False)
        self.assertIs(result.get("count"), None)
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("delete_identity_pool"),
        )
        self.assertEqual(len(mock_calls), 1)

    def _get_identity_pool_roles_side_effect(self, *args, **kwargs):
        if kwargs.get("IdentityPoolId") == first_pool_id:
            return first_pool_role_ret
        elif kwargs.get("IdentityPoolId") == third_pool_id:
            return third_pool_role_ret
        else:
            return default_pool_role_ret

    def test_that_when_get_identity_pool_roles_with_matching_pool_names_the_get_identity_pool_roles_method_returns_pools_role_properties(
        self,
    ):
        """
        Tests that the given 2 pool id's matching the given pool name, the results are
        passed through as a list of identity_pool_roles
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.get_identity_pool_roles.side_effect = (
            self._get_identity_pool_roles_side_effect
        )
        result = boto_cognitoidentity.get_identity_pool_roles(
            IdentityPoolName=first_pool_name, **conn_parameters
        )
        id_pool_roles = result.get("identity_pool_roles")
        self.assertIsNot(id_pool_roles, None)
        self.assertEqual(
            result.get("error", "key_should_not_be_there"), "key_should_not_be_there"
        )
        self.assertEqual(len(id_pool_roles), 2)
        self.assertEqual(id_pool_roles[0], first_pool_role_ret)
        self.assertEqual(id_pool_roles[1], third_pool_role_ret)

    def test_that_when_get_identity_pool_roles_with_no_matching_pool_names_the_get_identity_pool_roles_method_returns_none(
        self,
    ):
        """
        Tests that the given no pool id's matching the given pool name, the results returned is
        None and get_identity_pool_roles should never be called
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        result = boto_cognitoidentity.get_identity_pool_roles(
            IdentityPoolName="no_such_pool_name", **conn_parameters
        )
        mock_calls = self.conn.mock_calls
        self.assertIs(result.get("identity_pool_roles", "key_should_be_there"), None)
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(
            result.get("error", "key_should_not_be_there"), "key_should_not_be_there"
        )

    def test_that_when_get_identity_pool_roles_and_error_thrown_due_to_invalid_pool_id_the_get_identity_pool_roles_method_returns_error(
        self,
    ):
        """
        Tests that given an invalid pool id, we properly handle error generated from get_identity_pool_roles
        """
        self.conn.get_identity_pool_roles.side_effect = ClientError(
            error_content, "get_identity_pool_roles"
        )
        # try to delete an unexistent pool id "no_such_pool_id"
        result = boto_cognitoidentity.get_identity_pool_roles(
            IdentityPoolName="", IdentityPoolId="no_such_pool_id", **conn_parameters
        )
        mock_calls = self.conn.mock_calls
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("get_identity_pool_roles"),
        )
        self.assertEqual(len(mock_calls), 1)

    def test_that_when_set_identity_pool_roles_with_invalid_pool_id_the_set_identity_pool_roles_method_returns_set_false_and_error(
        self,
    ):
        """
        Tests that given an invalid pool id, we properly handle error generated from set_identity_pool_roles
        """
        self.conn.set_identity_pool_roles.side_effect = ClientError(
            error_content, "set_identity_pool_roles"
        )
        result = boto_cognitoidentity.set_identity_pool_roles(
            IdentityPoolId="no_such_pool_id", **conn_parameters
        )
        mock_calls = self.conn.mock_calls
        self.assertIs(result.get("set"), False)
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("set_identity_pool_roles"),
        )
        self.assertEqual(len(mock_calls), 1)

    def test_that_when_set_identity_pool_roles_with_no_roles_specified_the_set_identity_pool_roles_method_unset_the_roles(
        self,
    ):
        """
        Tests that given a valid pool id, and no other roles given, the role for the pool is cleared.
        """
        self.conn.set_identity_pool_roles.return_value = None
        result = boto_cognitoidentity.set_identity_pool_roles(
            IdentityPoolId="some_id", **conn_parameters
        )
        mock_calls = self.conn.mock_calls
        self.assertTrue(result.get("set"))
        self.assertEqual(len(mock_calls), 1)
        self.assertEqual(mock_calls[0][2].get("Roles"), {})

    def test_that_when_set_identity_pool_roles_with_only_auth_role_specified_the_set_identity_pool_roles_method_only_set_the_auth_role(
        self,
    ):
        """
        Tests that given a valid pool id, and only other given role is the AuthenticatedRole, the auth role for the
        pool is set and unauth role is cleared.
        """
        self.conn.set_identity_pool_roles.return_value = None
        with patch.dict(
            boto_cognitoidentity.__salt__,
            {
                "boto_iam.describe_role": MagicMock(
                    return_value={"arn": "my_auth_role_arn"}
                )
            },
        ):
            expected_roles = dict(authenticated="my_auth_role_arn")
            result = boto_cognitoidentity.set_identity_pool_roles(
                IdentityPoolId="some_id",
                AuthenticatedRole="my_auth_role",
                **conn_parameters
            )
            mock_calls = self.conn.mock_calls
            self.assertTrue(result.get("set"))
            self.assertEqual(len(mock_calls), 1)
            self.assertEqual(mock_calls[0][2].get("Roles"), expected_roles)

    def test_that_when_set_identity_pool_roles_with_only_unauth_role_specified_the_set_identity_pool_roles_method_only_set_the_unauth_role(
        self,
    ):
        """
        Tests that given a valid pool id, and only other given role is the UnauthenticatedRole, the unauth role for the
        pool is set and the auth role is cleared.
        """
        self.conn.set_identity_pool_roles.return_value = None
        with patch.dict(
            boto_cognitoidentity.__salt__,
            {
                "boto_iam.describe_role": MagicMock(
                    return_value={"arn": "my_unauth_role_arn"}
                )
            },
        ):
            expected_roles = dict(unauthenticated="my_unauth_role_arn")
            result = boto_cognitoidentity.set_identity_pool_roles(
                IdentityPoolId="some_id",
                UnauthenticatedRole="my_unauth_role",
                **conn_parameters
            )
            mock_calls = self.conn.mock_calls
            self.assertTrue(result.get("set"))
            self.assertEqual(len(mock_calls), 1)
            self.assertEqual(mock_calls[0][2].get("Roles"), expected_roles)

    def test_that_when_set_identity_pool_roles_with_both_roles_specified_the_set_identity_pool_role_method_set_both_roles(
        self,
    ):
        """
        Tests setting of both roles to valid given roles
        """
        self.conn.set_identity_pool_roles.return_value = None
        with patch.dict(
            boto_cognitoidentity.__salt__,
            {
                "boto_iam.describe_role": MagicMock(
                    return_value={"arn": "my_unauth_role_arn"}
                )
            },
        ):
            expected_roles = dict(
                authenticated="arn:aws:iam:my_auth_role",
                unauthenticated="my_unauth_role_arn",
            )
            result = boto_cognitoidentity.set_identity_pool_roles(
                IdentityPoolId="some_id",
                AuthenticatedRole="arn:aws:iam:my_auth_role",
                UnauthenticatedRole="my_unauth_role",
                **conn_parameters
            )
            mock_calls = self.conn.mock_calls
            self.assertTrue(result.get("set"))
            self.assertEqual(len(mock_calls), 1)
            self.assertEqual(mock_calls[0][2].get("Roles"), expected_roles)

    def test_that_when_set_identity_pool_roles_given_invalid_auth_role_the_set_identity_pool_method_returns_set_false_and_error(
        self,
    ):
        """
        Tests error handling for invalid auth role
        """
        with patch.dict(
            boto_cognitoidentity.__salt__,
            {"boto_iam.describe_role": MagicMock(return_value=False)},
        ):
            result = boto_cognitoidentity.set_identity_pool_roles(
                IdentityPoolId="some_id",
                AuthenticatedRole="no_such_auth_role",
                **conn_parameters
            )
            mock_calls = self.conn.mock_calls
            self.assertIs(result.get("set"), False)
            self.assertIn("no_such_auth_role", result.get("error", ""))
            self.assertEqual(len(mock_calls), 0)

    def test_that_when_set_identity_pool_roles_given_invalid_unauth_role_the_set_identity_pool_method_returns_set_false_and_error(
        self,
    ):
        """
        Tests error handling for invalid unauth role
        """
        with patch.dict(
            boto_cognitoidentity.__salt__,
            {"boto_iam.describe_role": MagicMock(return_value=False)},
        ):
            result = boto_cognitoidentity.set_identity_pool_roles(
                IdentityPoolId="some_id",
                AuthenticatedRole="arn:aws:iam:my_auth_role",
                UnauthenticatedRole="no_such_unauth_role",
                **conn_parameters
            )
            mock_calls = self.conn.mock_calls
            self.assertIs(result.get("set"), False)
            self.assertIn("no_such_unauth_role", result.get("error", ""))
            self.assertEqual(len(mock_calls), 0)

    def test_that_when_update_identity_pool_given_invalid_pool_id_the_update_identity_pool_method_returns_updated_false_and_error(
        self,
    ):
        """
        Tests error handling for invalid pool id
        """
        self.conn.describe_identity_pool.side_effect = ClientError(
            error_content, "error on describe identity pool with pool id"
        )
        result = boto_cognitoidentity.update_identity_pool(
            IdentityPoolId="no_such_pool_id", **conn_parameters
        )
        self.assertIs(result.get("updated"), False)
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("error on describe identity pool with pool id"),
        )

    def test_that_when_update_identity_pool_given_only_valid_pool_id_the_update_identity_pool_method_returns_udpated_identity(
        self,
    ):
        """
        Tests the base case of calling update_identity_pool with only the pool id, verify
        that the passed parameters into boto3 update_identity_pool has at least all the
        expected required parameters in IdentityPoolId, IdentityPoolName and AllowUnauthenticatedIdentities
        """
        self.conn.describe_identity_pool.return_value = second_pool_ret
        self.conn.update_identity_pool.return_value = second_pool_ret
        expected_params = second_pool_ret
        result = boto_cognitoidentity.update_identity_pool(
            IdentityPoolId=second_pool_id, **conn_parameters
        )
        self.assertTrue(result.get("updated"))
        self.assertEqual(result.get("identity_pool"), second_pool_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_valid_pool_id_and_pool_name_the_update_identity_pool_method_returns_updated_identity_pool(
        self,
    ):
        """
        Tests successful update of a pool's name
        """
        self.conn.describe_identity_pool.return_value = second_pool_ret
        second_pool_updated_ret = second_pool_ret.copy()
        second_pool_updated_ret["IdentityPoolName"] = second_pool_name_updated
        self.conn.update_identity_pool.return_value = second_pool_updated_ret
        expected_params = second_pool_updated_ret
        result = boto_cognitoidentity.update_identity_pool(
            IdentityPoolId=second_pool_id,
            IdentityPoolName=second_pool_name_updated,
            **conn_parameters
        )
        self.assertTrue(result.get("updated"))
        self.assertEqual(result.get("identity_pool"), second_pool_updated_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_empty_dictionary_for_supported_login_providers_the_update_identity_pool_method_is_called_with_proper_request_params(
        self,
    ):
        """
        Tests the request parameters to boto3 update_identity_pool's AllowUnauthenticatedIdentities is {}
        """
        self.conn.describe_identity_pool.return_value = first_pool_ret
        first_pool_updated_ret = first_pool_ret.copy()
        first_pool_updated_ret.pop("SupportedLoginProviders")
        self.conn.update_identity_pool.return_value = first_pool_updated_ret
        expected_params = first_pool_ret.copy()
        expected_params["SupportedLoginProviders"] = {}
        expected_params.pop("DeveloperProviderName")
        expected_params.pop("OpenIdConnectProviderARNs")
        result = boto_cognitoidentity.update_identity_pool(
            IdentityPoolId=first_pool_id, SupportedLoginProviders={}, **conn_parameters
        )
        self.assertTrue(result.get("updated"))
        self.assertEqual(result.get("identity_pool"), first_pool_updated_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_empty_list_for_openid_connect_provider_arns_the_update_identity_pool_method_is_called_with_proper_request_params(
        self,
    ):
        """
        Tests the request parameters to boto3 update_identity_pool's OpenIdConnectProviderARNs is []
        """
        self.conn.describe_identity_pool.return_value = first_pool_ret
        first_pool_updated_ret = first_pool_ret.copy()
        first_pool_updated_ret.pop("OpenIdConnectProviderARNs")
        self.conn.update_identity_pool.return_value = first_pool_updated_ret
        expected_params = first_pool_ret.copy()
        expected_params.pop("SupportedLoginProviders")
        expected_params.pop("DeveloperProviderName")
        expected_params["OpenIdConnectProviderARNs"] = []
        result = boto_cognitoidentity.update_identity_pool(
            IdentityPoolId=first_pool_id,
            OpenIdConnectProviderARNs=[],
            **conn_parameters
        )
        self.assertTrue(result.get("updated"))
        self.assertEqual(result.get("identity_pool"), first_pool_updated_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_developer_provider_name_when_developer_provider_name_was_set_previously_the_udpate_identity_pool_method_is_called_without_developer_provider_name_param(
        self,
    ):
        """
        Tests the request parameters do not include 'DeveloperProviderName' if this was previously set
        for the given pool id
        """
        self.conn.describe_identity_pool.return_value = first_pool_ret
        self.conn.update_identity_pool.return_value = first_pool_ret
        expected_params = first_pool_ret.copy()
        expected_params.pop("SupportedLoginProviders")
        expected_params.pop("DeveloperProviderName")
        expected_params.pop("OpenIdConnectProviderARNs")
        result = boto_cognitoidentity.update_identity_pool(
            IdentityPoolId=first_pool_id,
            DeveloperProviderName="this should not change",
            **conn_parameters
        )
        self.assertTrue(result.get("updated"))
        self.assertEqual(result.get("identity_pool"), first_pool_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_when_update_identity_pool_given_developer_provider_name_is_included_in_the_params_when_associated_for_the_first_time(
        self,
    ):
        """
        Tests the request parameters include 'DeveloperProviderName' when the pool did not have this
        property set previously.
        """
        self.conn.describe_identity_pool.return_value = second_pool_ret
        second_pool_updated_ret = second_pool_ret.copy()
        second_pool_updated_ret["DeveloperProviderName"] = "added_developer_provider"
        self.conn.update_identity_pool.return_value = second_pool_updated_ret
        expected_params = second_pool_updated_ret.copy()
        result = boto_cognitoidentity.update_identity_pool(
            IdentityPoolId=second_pool_id,
            DeveloperProviderName="added_developer_provider",
            **conn_parameters
        )
        self.assertTrue(result.get("updated"))
        self.assertEqual(result.get("identity_pool"), second_pool_updated_ret)
        self.conn.update_identity_pool.assert_called_with(**expected_params)

    def test_that_the_update_identity_pool_method_handles_exception_from_boto3(self):
        """
        Tests the error handling of exception generated by boto3 update_identity_pool
        """
        self.conn.describe_identity_pool.return_value = second_pool_ret
        second_pool_updated_ret = second_pool_ret.copy()
        second_pool_updated_ret["DeveloperProviderName"] = "added_developer_provider"
        self.conn.update_identity_pool.side_effect = ClientError(
            error_content, "update_identity_pool"
        )
        result = boto_cognitoidentity.update_identity_pool(
            IdentityPoolId=second_pool_id,
            DeveloperProviderName="added_developer_provider",
            **conn_parameters
        )
        self.assertIs(result.get("updated"), False)
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("update_identity_pool"),
        )



import shutil

import pytest

from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing(*KNOWN_BINARY_NAMES, check_all=False),
]


@pytest.fixture
def venv_dir(tmp_path):
    return tmp_path / "venv"


@pytest.fixture
def virtualenv(modules):
    return modules.virtualenv


def test_create_defaults(virtualenv, venv_dir):
    """
    virtualenv.managed
    """
    ret = virtualenv.create(str(venv_dir))
    assert ret
    assert "retcode" in ret
    assert ret["retcode"] == 0
    pip_binary = venv_dir / "bin" / "pip"
    assert pip_binary.exists()


def test_site_packages(virtualenv, venv_dir, modules):
    ret = virtualenv.create(str(venv_dir), system_site_packages=True)
    assert ret
    assert "retcode" in ret
    assert ret["retcode"] == 0
    pip_binary = venv_dir / "bin" / "pip"
    with_site = modules.pip.freeze(bin_env=str(pip_binary))
    shutil.rmtree(venv_dir)
    ret = virtualenv.create(str(venv_dir))
    assert ret
    assert "retcode" in ret
    assert ret["retcode"] == 0
    without_site = modules.pip.freeze(bin_env=str(pip_binary))
    assert with_site != without_site


def test_clear(virtualenv, venv_dir, modules):
    ret = virtualenv.create(str(venv_dir))
    assert ret
    assert "retcode" in ret
    assert ret["retcode"] == 0
    pip_binary = venv_dir / "bin" / "pip"
    modules.pip.install("pep8", bin_env=str(pip_binary))
    ret = virtualenv.create(str(venv_dir), clear=True)
    assert ret
    assert "retcode" in ret
    assert ret["retcode"] == 0
    packages = modules.pip.list(prefix="pep8", bin_env=str(pip_binary))
    assert "pep8" not in packages


def test_virtualenv_ver(virtualenv, venv_dir):
    ret = virtualenv.create(str(venv_dir))
    assert ret
    assert "retcode" in ret
    assert ret["retcode"] == 0
    ret = virtualenv.virtualenv_ver(str(venv_dir))
    assert isinstance(ret, tuple)
    assert all([isinstance(x, int) for x in ret])



"""
    :codeauthor: :email:`Gareth J. Greenaway <gareth@saltstack.com>`
    :codeauthor: :email:`David Murphy <dmurphy@saltstack.com>`
"""

import datetime
import logging
import pathlib
import shutil
import subprocess
import time
import types

import psutil
import pytest

import salt.modules.gpg as gpg
from tests.support.mock import MagicMock, Mock, call, patch

pytest.importorskip("gnupg")

pytestmark = [
    pytest.mark.skip_unless_on_linux,
    pytest.mark.requires_random_entropy,
    pytest.mark.slow_test,
]

log = logging.getLogger(__name__)


GPG_TEST_KEY_PASSPHRASE = "testkeypassphrase"
GPG_TEST_KEY_ID = "7416F045"
GPG_TEST_PUB_KEY = """-----BEGIN PGP PUBLIC KEY BLOCK-----

mQGNBFz1dx4BDACph7J5nuWE+zb9rZqTaL8akAnPAli2j6Qtk7BTDzTM9Kq80U2P
O3QRAFBQDODsgmrBTWgeZeNhN6Snz2WrZ8pC0RMK+mCmEt5S49ydWtvWl/xtzPfg
sy8h8OrIjSH1G0NQv9vdBpg2Y9vXLNDCZAqH0/Sh/wCkHT80c4i8TU09UZfx96S6
fFVmB7nHu/ztgiIkC6Fr04WGHlkd50g8r8CFPvwKOTD+rfoMsGppmAC1+y8ajfik
B+tEL88Rv2s4hLU78nQ3iGtdVRSs5Ip0x4O/PYZIDEd8KuBMo+flSzOZj2HCbUzN
MHiTXIbL8hBlwku9LGO7Itks4v2vfDh57JRHykwzGnvOlgXtSSvbayWLohNXtzWq
WtsMKrsrsIUprg1jhIVHnMSZZXMMizlni6MT5vbil2Bgn1g7diNscDnfCD6vlWUH
FRS1XkFFZ5ozu0+lC/5UaTcrjhH8/hQBNJwjsDSPsyKYQm+t7PXqq4SlI3Er7JJW
esK0diZ6reeebl0AEQEAAbQhdGVzdGtleSA8cGFja2FnaW5nQHNhbHRzdGFjay5j
b20+iQHOBBMBCgA4AhsDBQsJCAcCBhUKCQgLAgQWAgMBAh4BAheAFiEEjS1DixNC
naoZrRFuvMeUg3QW8EUFAlz1ekoACgkQvMeUg3QW8EVm1gv/Z5CCqFoF8IXpa3g0
G+9C4gRS0h+tEYtjebAgxBn/dqk6cSNIb1BGDM/sSWxK5/jweN8CF/ojVgP1CFrX
igULnL3g5351L+8feU2ImP2OML438/GcOOJ+iwTKUBRuqxqK8NwcRuuN6YmbjtUw
JSD2+60DlWfGsInyenwGkgBAM44Y6dSAXKsm6dJ/IGVTQIsHE5mKRykrRZ1jyXxw
i1CF8fEyG6fNNb8I8Oiyj52xdYMxBvGmgPpFlF1jRoU+OViGFeUiZ49XFlC8GNCf
boWfXg+PCwZeWUeJ/s6a6iC5HG7iC0XYRiqhzymP8ctrwwm5dPT4ebYNfUSqq8i0
/YG6JGdsQZrq0i0FuceUSfOKVeFvSm+AkCG3kLxxgM925fqBetWiAJRCDYdmy8uc
jm618584hqa/yFMOfwVoLqq8mVg84wyGU6gXo6hXCa/008PITFZD47ILuN/Z8iLm
8Or/aJD5Tq+Wg3r2Di9+Ku8/gcn3bVxwFRjiVPrFwJkyjfMFuQGNBFz1dx4BDADU
Ynz0efLmaKqZIjfxwDf7iVqwcD9b8cMXIv2aa3Y56QDVJBXU5952LiU8XuzBBPq+
4FYQ78UmxSv3Rk6LKb9P2ih2L1PaJuQ1ZkNrQLqab3olpAu/Xe3raGLgCOU0RKJw
EPF3RcKu8ALuRcovfwzXWg8w19QRUPewZdVC4VgslKp8mNLECvdUxxVIDQWf06RZ
uCAfbqdiYReE62QT7NR4lAa1RpfU7Nt149OcQEP8VKTAZgTYyuwhRXFbrDD3Zp58
k5H0nKHNX+w1Ayih/YUk2b3etaBhlcTVAy/73TPfrd3Gl8dtzJZNtUD/eLWdGfP9
mCghmyAqbiQngH2eAMeifIYornynZFVBPBlvnwy7Iouq0V6tIVyNPGp0jcy1j2XT
NRBJyFbvam3hmrRW8A/VOJQ1W7LOKaM/5lh/BarrSLKn0xlL97GTmuSqlS+WrmyM
kU182TUYyUD7Rs3mydnMVS/N4aRxu4ITaTm9vieZLmAPR9vPgo+GwdHEkwm797kA
EQEAAYkBtgQYAQoAIAIbDBYhBI0tQ4sTQp2qGa0RbrzHlIN0FvBFBQJc9XqkAAoJ
ELzHlIN0FvBFlyEL/jVhm2PFj2mCLuKE5/nV4JvxY4Qu4+NCFiEdYK+zUoD36gEJ
3VjHL5dydHuZWcvm+XLW1PseNx2awVs47mjv2iZOLwY6BtfAFWhWEFmBEe6fTFXz
KkDWRst4gm0b0B7S3byoABwcyYNS6RkTfUApK4zdYErbfOLoT+Xa08YQKLVK7fmE
KBnBnnHUvktYTEvhwv9BID+qLnTVSQcjRcXbDQAYm14c7Nyb/SyxcUaUkCk41MVY
+vzNQlFrVc4h2np41X8JbmrsQb37E7lE+h32sJFBU03SGf0vT7SXXQj+UD/DEGay
Gt/8aRa5FGrcJyM5mTdbSgvCp0EjTrdokK5GHwM23cbSTo+nN9BNhIBRc4929SaJ
DVRqOIoJ+eHZdf3gIkMPOA3fBbMYzW65LIxt/p49tHD0c/nioZETycEgGuuYbnrn
IfXHFqiCAxkobIHqUg/BSu1cs8GNgE7BVUXy8JThuzmVdh4Pvd3YN1ouoPyVuDrk
ylirh0aqUQdSeIuJTg==
=yF8M
-----END PGP PUBLIC KEY BLOCK-----
"""

GPG_TEST_PRIV_KEY = """-----BEGIN PGP PRIVATE KEY BLOCK-----


OZV2Hg+93dg3Wi6g/JW4OuTKWKuHRqpRB1J4i4lO
=WRTN
-----END PGP PRIVATE KEY BLOCK-----
"""


@pytest.fixture
def gpghome(tmp_path):
    root = tmp_path / "gpghome"
    root.mkdir(mode=0o0700)
    root.joinpath(".gnupg").mkdir(mode=0o0700)
    pub = root / "gpgfile.pub"
    pub.write_text(GPG_TEST_PUB_KEY)
    priv = root / "gpgfile.priv"
    priv.write_text(GPG_TEST_PRIV_KEY)
    try:
        yield types.SimpleNamespace(path=root, pub=pub, priv=priv)
    finally:
        # Make sure we don't leave any gpg-agent's running behind
        gpg_connect_agent = shutil.which("gpg-connect-agent")
        if gpg_connect_agent:
            gnupghome = root / ".gnupg"
            if not gnupghome.is_dir():
                gnupghome = root
            try:
                subprocess.run(
                    [gpg_connect_agent, "killagent", "/bye"],
                    env={"GNUPGHOME": str(gnupghome)},
                    shell=False,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except subprocess.CalledProcessError:
                # This is likely CentOS 7 or Amazon Linux 2
                pass

        # If the above errored or was not enough, as a last resort, let's check
        # the running processes.
        for proc in psutil.process_iter():
            try:
                if "gpg-agent" in proc.name():
                    for arg in proc.cmdline():
                        if str(root) in arg:
                            proc.terminate()
            except Exception:  # pylint: disable=broad-except
                pass


@pytest.fixture
def configure_loader_modules(gpghome):
    return {gpg: {}}


def test_list_keys():
    """
    Test gpg.list_keys
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": "/root",
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        },
        {
            "dummy": "",
            "keyid": "yyyyyyyyyyyyyyyy",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                [
                    "yyyyyyyyyyyyyyyy",
                    "e",
                    "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy",
                ]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy",
            "date": "1506612692",
            "trust": "r",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        },
    ]

    _expected_result = [
        {
            "keyid": "xxxxxxxxxxxxxxxx",
            "uids": ["GPG Person <person@example.com>"],
            "created": "2017-09-28",
            "expires": "2033-09-24",
            "keyLength": "4096",
            "ownerTrust": "Unknown",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "trust": "Unknown",
        },
        {
            "keyid": "yyyyyyyyyyyyyyyy",
            "uids": ["GPG Person <person@example.com>"],
            "created": "2017-09-28",
            "expires": "2033-09-24",
            "keyLength": "4096",
            "ownerTrust": "Unknown",
            "fingerprint": "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy",
            "trust": "Revoked",
        },
    ]

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                assert gpg.list_keys() == _expected_result


def test_get_key():
    """
    Test gpg.get_key
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": "/root",
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        }
    ]

    _expected_result = {
        "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "keyid": "xxxxxxxxxxxxxxxx",
        "uids": ["GPG Person <person@example.com>"],
        "created": "2017-09-28",
        "trust": "Unknown",
        "ownerTrust": "Unknown",
        "expires": "2033-09-24",
        "keyLength": "4096",
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                ret = gpg.get_key("xxxxxxxxxxxxxxxx")
                assert ret == _expected_result


def test_delete_key_without_passphrase(gpghome):
    """
    Test gpg.delete_key without passphrase
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        }
    ]

    _expected_result = {
        "res": True,
        "message": (
            "Secret key for xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx deleted\nPublic"
            " key for xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx deleted"
        ),
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                with patch(
                    "salt.modules.gpg.gnupg.GPG.delete_keys",
                    MagicMock(return_value="ok"),
                ) as gnupg_delete_keys:
                    ret = gpg.delete_key(
                        "xxxxxxxxxxxxxxxx", delete_secret=True, use_passphrase=False
                    )
                    assert ret == _expected_result
                    gnupg_delete_keys.assert_called_with(
                        "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
                        False,
                        expect_passphrase=False,
                    )


def test_delete_key_with_passphrase_without_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.delete_key with passphrase but without gpg_passphrase pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        }
    ]

    _expected_result = {
        "res": False,
        "message": "Failed to delete secret key for xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx: gpg_passphrase not available in pillar.",
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=None)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                with patch(
                    "salt.modules.gpg.gnupg.GPG.delete_keys"
                ) as gnupg_delete_keys:
                    ret = gpg.delete_key("xxxxxxxxxxxxxxxx", delete_secret=True)
                    assert ret == _expected_result
                    gnupg_delete_keys.assert_not_called()


def test_delete_key_with_passphrase_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.delete_key with passphrase and gpg_passphrase pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _list_result = [
        {
            "dummy": "",
            "keyid": "xxxxxxxxxxxxxxxx",
            "expires": "2011188692",
            "sigs": [],
            "subkeys": [
                ["xxxxxxxxxxxxxxxx", "e", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"]
            ],
            "length": "4096",
            "ownertrust": "-",
            "sig": "",
            "algo": "1",
            "fingerprint": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "date": "1506612692",
            "trust": "-",
            "type": "pub",
            "uids": ["GPG Person <person@example.com>"],
        }
    ]

    _expected_result = {
        "res": True,
        "message": (
            "Secret key for xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx deleted\nPublic"
            " key for xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx deleted"
        ),
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            with patch.object(gpg, "_list_keys", return_value=_list_result):
                with patch(
                    "salt.modules.gpg.gnupg.GPG.delete_keys",
                    MagicMock(return_value="ok"),
                ) as gnupg_delete_keys:
                    ret = gpg.delete_key("xxxxxxxxxxxxxxxx", delete_secret=True)
                    assert ret == _expected_result
                    gnupg_delete_keys.assert_any_call(
                        "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
                        True,
                        passphrase=GPG_TEST_KEY_PASSPHRASE,
                    )
                    gnupg_delete_keys.assert_called_with(
                        "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
                        False,
                        expect_passphrase=False,
                    )


def test_export_public_key(gpghome):
    """
    Test gpg.export_key with single public key
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PUB_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key("xxxxxxxxxxxxxxxx", bare=True)
                assert ret == GPG_TEST_PUB_KEY
                gnupg_export_keys.assert_called_with(
                    ["xxxxxxxxxxxxxxxx"],
                    False,
                    expect_passphrase=False,
                )


def test_export_public_key_to_file(gpghome):
    """
    Test gpg.export_key output public key to file
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    exported_keyfile = gpghome.path / "exported_pub_key"
    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)

    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PUB_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key(
                    keyids="xxxxxxxxxxxxxxxx", output=exported_keyfile, bare=True
                )
                assert ret == GPG_TEST_PUB_KEY
                keyfile_contents = pathlib.Path(exported_keyfile).read_text(
                    encoding="utf-8"
                )
                assert keyfile_contents == GPG_TEST_PUB_KEY


def test_export_multiple_public_keys(gpghome):
    """
    Test gpg.export_key with multiple public keys
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PUB_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key(
                    "xxxxxxxxxxxxxxxx,yyyyyyyyyyyyyyyy,zzzzzzzzzzzzzzzz",
                    bare=True,
                )
                assert ret == GPG_TEST_PUB_KEY
                gnupg_export_keys.assert_called_with(
                    ["xxxxxxxxxxxxxxxx", "yyyyyyyyyyyyyyyy", "zzzzzzzzzzzzzzzz"],
                    False,
                    expect_passphrase=False,
                )


def test_export_secret_key_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.export_key with secret key and gpg_passphrase in pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PRIV_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key(
                    "xxxxxxxxxxxxxxxx", secret=True, use_passphrase=True, bare=True
                )
                assert ret == GPG_TEST_PRIV_KEY
                gnupg_export_keys.assert_called_with(
                    ["xxxxxxxxxxxxxxxx"],
                    True,
                    passphrase=GPG_TEST_KEY_PASSPHRASE,
                )


def test_export_secret_key_to_file_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.export_key output secret key to file with gpg_passphrase in pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    exported_keyfile = gpghome.path / "exported_priv_key"
    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            with patch(
                "salt.modules.gpg.gnupg.GPG.export_keys",
                MagicMock(return_value=GPG_TEST_PRIV_KEY),
            ) as gnupg_export_keys:
                ret = gpg.export_key(
                    keyids="xxxxxxxxxxxxxxxx",
                    secret=True,
                    output=exported_keyfile,
                    use_passphrase=True,
                    bare=True,
                )
                assert ret == GPG_TEST_PRIV_KEY
                gnupg_export_keys.assert_called_with(
                    ["xxxxxxxxxxxxxxxx"],
                    True,
                    passphrase=GPG_TEST_KEY_PASSPHRASE,
                )
                keyfile_contents = pathlib.Path(exported_keyfile).read_text(
                    encoding="utf-8"
                )
                assert keyfile_contents == GPG_TEST_PRIV_KEY


def test_create_key_without_passphrase(gpghome):
    """
    Test gpg.create_key without a passphrase
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            ret = gpg.create_key()
            assert ret["res"]
            assert ret["fingerprint"]
            assert ret["message"] == "GPG key pair successfully generated."


def test_create_key_with_passphrase_without_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.create_key with passphrase but without gpg_passphrase pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=None)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            ret = gpg.create_key(use_passphrase=True)
            assert not ret["res"]
            assert ret["message"] == "gpg_passphrase not available in pillar."


def test_create_key_with_passphrase_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.create_key with passphrase and gpg_passphrase pillar
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    mock_opt = MagicMock(return_value="root")
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}), patch.dict(
            gpg.__salt__, {"pillar.get": pillar_mock}
        ):
            ret = gpg.create_key(use_passphrase=True)
            assert ret["res"]
            assert ret["fingerprint"]
            assert ret["message"] == "GPG key pair successfully generated."


def test_search_keys(gpghome):
    """
    Test gpg.search_keys
    """

    _user_mock = {
        "shell": "/bin/bash",
        "workphone": "",
        "uid": 0,
        "passwd": "x",
        "roomnumber": "",
        "gid": 0,
        "groups": ["root"],
        "home": str(gpghome.path),
        "fullname": "root",
        "homephone": "",
        "name": "root",
    }

    _search_result = [
        {
            "keyid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "uids": ["GPG Person <person@example.com>"],
            "expires": "",
            "sigs": [],
            "length": "1024",
            "algo": "17",
            "date": int(time.mktime(datetime.datetime(2004, 11, 13).timetuple())),
            "type": "pub",
        }
    ]

    _expected_result = [
        {
            "uids": ["GPG Person <person@example.com>"],
            "created": "2004-11-13",
            "keyLength": "1024",
            "keyid": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        }
    ]

    mock_search_keys = MagicMock(return_value=_search_result)
    mock_opt = MagicMock(return_value="root")
    with patch.dict(gpg.__salt__, {"user.info": MagicMock(return_value=_user_mock)}):
        with patch.dict(gpg.__salt__, {"config.option": mock_opt}):
            with patch.object(gpg, "_search_keys", mock_search_keys):
                ret = gpg.search_keys("person@example.com")
                assert ret == _expected_result

                assert (
                    call(
                        "person@example.com",
                        "keys.openpgp.org",
                        user=None,
                        gnupghome=None,
                    )
                    in mock_search_keys.mock_calls
                )

                ret = gpg.search_keys("person@example.com", "keyserver.ubuntu.com")
                assert ret == _expected_result

                assert (
                    call(
                        "person@example.com",
                        "keyserver.ubuntu.com",
                        user=None,
                        gnupghome=None,
                    )
                    in mock_search_keys.mock_calls
                )


def test_gpg_import_pub_key(gpghome):
    config_user = MagicMock(return_value="salt")
    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            ret = gpg.import_key(None, str(gpghome.pub), "salt", str(gpghome.path))
            assert ret["res"] is True


def test_gpg_import_priv_key(gpghome):
    config_user = MagicMock(return_value="salt")
    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            ret = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
            assert ret["res"] is True


def test_gpg_sign(gpghome):
    """
    Test gpg.sign
    """

    config_user = MagicMock(return_value="salt")
    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            with patch.dict(gpg.__salt__, {"pillar.get": pillar_mock}):
                key = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
                assert key["res"] is True
                gpg_text_input = "The quick brown fox jumped over the lazy dog"
                ret = gpg.sign(
                    keyid=GPG_TEST_KEY_ID,
                    text=gpg_text_input,
                    use_passphrase=True,
                    gnupghome=str(gpghome.path),
                )
                assert "-----BEGIN PGP SIGNATURE-----" in str(ret)


def test_gpg_encrypt_message(gpghome):
    """
    Test gpg.encrypt
    """

    config_user = MagicMock(return_value="salt")

    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            key = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
            assert key["res"] is True
            gpg_text_input = "The quick brown fox jumped over the lazy dog"
            ret = gpg.encrypt(
                recipients=GPG_TEST_KEY_ID,
                text=gpg_text_input,
                use_passphrase=True,
                always_trust=True,
                gnupghome=str(gpghome.path),
            )
            assert ret["res"] is True


def test_gpg_encrypt_and_sign_message_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.encrypt sign message with passphrase and gpg_passphrase in pillar
    """

    config_user = MagicMock(return_value="salt")

    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            with patch.dict(gpg.__salt__, {"pillar.get": pillar_mock}):
                key = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
                assert key["res"] is True
                gpg_text_input = "The quick brown fox jumped over the lazy dog"
                ret = gpg.encrypt(
                    recipients=GPG_TEST_KEY_ID,
                    text=gpg_text_input,
                    sign=True,
                    use_passphrase=True,
                    always_trust=True,
                    gnupghome=str(gpghome.path),
                )
                assert ret["res"] is True


def test_gpg_decrypt_message_with_gpg_passphrase_in_pillar(gpghome):
    """
    Test gpg.decrypt with passphrase and gpg_passphrase in pillar
    """

    gpg_encrypted_message = """-----BEGIN PGP MESSAGE-----
    hQGMA7z9rKs9ZvTOAQwAnMbwchCm1VXOD+Ml0rnNrhDhsRm+6O96FOq5lWY0ntkj
    vnXeFOgUf0wzK4hkQT/Yo4/ZpDkV3iwwSIjesqNDS1U/KWfbe2pFeph6w9fHFnXf
    e8RJY8OGJHN8A9TlCIyKDVKGoXZEcSDZ4K0fNL/OxF3LxsSjJ894F+e2MCnoQZgq
    3uxQFYb9Gec8iPpBcd7s74J8cX1To9rBKRNSP73NRwVvKbPgB0NjzrLVPmqHGO7Q
    wFf9tZdGYmxV/AB9H0gCUITBCwnndissVqFpjUpVEIL1jOBus73oiL9Grn5QlZic
    wDOZyKDlWR15jiiDaWvq4ES6O0j2Oj/FodItJhTbzeXEFOxn7cNbW2kAAc08Hs1u
    zPI2LywaCSM/IOxrRI8n/ExXxtU0k48827AWTLYmhm+vDct7gQHEJ8Qu3/d5FsP/
    f1qWrtl+E7TpIIlUIgtPL596LYiwQLisqh1JeqqleSnBESIzgg7teolYdYED606z
    j9a1oTUHvSnKz6Q5Yh850mYBCBc/iEn0gJBgseGK6rXJF54f5duybkuTKrV+7hUd
    euy/+K0mnQWPS0G9NhCdQAaydNBZywOyQU6Yv1gPHapRrHOkhZq51sBKmZRGdqH/
    llX+Wd7pUetBV+vmZp75l8q9CNVpefA=
    =WXYF
    -----END PGP MESSAGE-----
    """

    config_user = MagicMock(return_value="salt")
    user_info = MagicMock(
        return_value={"name": "salt", "home": str(gpghome.path), "uid": 1000}
    )
    pillar_mock = MagicMock(return_value=GPG_TEST_KEY_PASSPHRASE)
    with patch.dict(gpg.__salt__, {"config.option": config_user}):
        with patch.dict(gpg.__salt__, {"user.info": user_info}):
            with patch.dict(gpg.__salt__, {"pillar.get": pillar_mock}):
                key = gpg.import_key(None, str(gpghome.priv), "salt", str(gpghome.path))
                assert key["res"] is True
                ret = gpg.decrypt(
                    text=gpg_encrypted_message,
                    use_passphrase=True,
                    gnupghome=str(gpghome.path),
                )
                assert ret["res"] is True


@pytest.fixture(params={})
def _import_result_mock(request):
    defaults = {
        "gpg": Mock(),
        "imported": 0,
        "results": [],
        "fingerprints": [],
        "count": 0,
        "no_user_id": 0,
        "imported_rsa": 0,
        "unchanged": 0,
        "n_uids": 0,
        "n_subk": 0,
        "n_sigs": 0,
        "n_revoc": 0,
        "sec_read": 0,
        "sec_imported": 0,
        "sec_dups": 0,
        "not_imported": 0,
        "stderr": "",
        "data": b"",
    }
    defaults.update(request.param)
    import_result = MagicMock()
    import_result.__bool__.return_value = False
    for var, val in defaults.items():
        setattr(import_result, var, val)
    return import_result


@pytest.mark.parametrize(
    "_import_result_mock",
    (
        {
            "count": 1,
            "stderr": "gpg: key ABCDEF0123456789: no user ID\ngpg: Total number processed: 1\n[GNUPG:] IMPORT_RES 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0\n",
        },
    ),
    indirect=True,
)
def test_gpg_receive_keys_no_user_id(_import_result_mock):
    with patch("salt.modules.gpg._create_gpg") as create:
        with patch.dict(
            gpg.__salt__, {"user.info": MagicMock(), "config.option": Mock()}
        ):
            create.return_value.recv_keys.return_value = _import_result_mock
            res = gpg.receive_keys(keys="abc", user="abc")
            assert res["res"] is False
            assert any("no user ID" in x for x in res["message"])


@pytest.mark.parametrize(
    "_import_result_mock",
    (
        {
            "results": [{"fingerprint": None, "problem": "0", "text": "Other failure"}],
            "stderr": "[GNUPG:] FAILURE recv-keys 167772346\ngpg: keyserver receive failed: No keyserver available\n",
            "returncode": 2,
        },
    ),
    indirect=True,
)
def test_gpg_receive_keys_keyserver_unavailable(_import_result_mock):
    with patch("salt.modules.gpg._create_gpg") as create:
        with patch.dict(
            gpg.__salt__, {"user.info": MagicMock(), "config.option": Mock()}
        ):
            create.return_value.recv_keys.return_value = _import_result_mock
            res = gpg.receive_keys(keys="abc", user="abc")
            assert res["res"] is False
            assert any("No keyserver available" in x for x in res["message"])



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import os

import pytest

import salt.modules.environ as environ
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {environ: {}}


def test_setval():
    """
    Test for set a single salt process environment variable. Returns True
    on success.
    """
    mock = MagicMock(return_value=None)
    with patch.dict(os.environ, {}):
        assert environ.setval("key", False, True) is None

    mock = MagicMock(side_effect=Exception())
    with patch.dict(os.environ, {}):
        assert not environ.setval("key", False, True)

    mock_environ = {}
    with patch.dict(os.environ, mock_environ):
        assert environ.setval("key", False) == ""

    with patch.dict(os.environ, mock_environ):
        assert not environ.setval("key", True)


def test_set_val_permanent():
    with patch.dict(os.environ, {}), patch.dict(
        environ.__utils__,
        {"reg.set_value": MagicMock(), "reg.delete_value": MagicMock()},
    ), patch("salt.utils.platform.is_windows", return_value=True):

        environ.setval("key", "Test", permanent=True)
        environ.__utils__["reg.set_value"].assert_called_with(
            "HKCU", "Environment", "key", "Test"
        )


def test_set_val_permanent_false_unsets():
    with patch.dict(os.environ, {}), patch.dict(
        environ.__utils__,
        {"reg.set_value": MagicMock(), "reg.delete_value": MagicMock()},
    ), patch("salt.utils.platform.is_windows", return_value=True):

        environ.setval("key", False, false_unsets=True, permanent=True)
        environ.__utils__["reg.set_value"].assert_not_called()
        environ.__utils__["reg.delete_value"].assert_called_with(
            "HKCU", "Environment", "key"
        )


def test_set_val_permanent_hklm():
    with patch.dict(os.environ, {}), patch.dict(
        environ.__utils__,
        {"reg.set_value": MagicMock(), "reg.delete_value": MagicMock()},
    ), patch("salt.utils.platform.is_windows", return_value=True):

        key = r"SYSTEM\CurrentControlSet\Control\Session Manager\Environment"
        environ.setval("key", "Test", permanent="HKLM")
        environ.__utils__["reg.set_value"].assert_called_with(
            "HKLM", key, "key", "Test"
        )


def test_setenv():
    """
    Set multiple salt process environment variables from a dict.
    Returns a dict.
    """
    mock_environ = {"KEY": "value"}
    with patch.dict(os.environ, mock_environ):
        assert not environ.setenv("environ")

    with patch.dict(os.environ, mock_environ):
        assert not environ.setenv({"A": True}, False, True, False)

    with patch.dict(os.environ, mock_environ):
        mock_setval = MagicMock(return_value=None)
        with patch.object(environ, "setval", mock_setval):
            assert environ.setenv({}, False, True, False)["KEY"] is None


def test_get():
    """
    Get a single salt process environment variable.
    """
    assert not environ.get(True)

    assert environ.get("key") == ""


def test_has_value():
    """
    Determine whether the key exists in the current salt process
    environment dictionary. Optionally compare the current value
    of the environment against the supplied value string.
    """
    mock_environ = {}
    with patch.dict(os.environ, mock_environ):
        assert not environ.has_value(True)

        os.environ["salty"] = "yes"
        assert environ.has_value("salty", "yes")

        os.environ["too_salty"] = "no"
        assert not environ.has_value("too_salty", "yes")

        assert not environ.has_value("key", "value")

        os.environ["key"] = "value"
        assert environ.has_value("key")


def test_item():
    """
    Get one or more salt process environment variables.
    Returns a dict.
    """
    assert environ.item(None) == {}


def test_items():
    """
    Return a dict of the entire environment set for the salt process
    """
    assert list(environ.items()) != []



import logging
import socket
import textwrap

import pytest

import salt.modules.cmdmod
import salt.utils.dns
from salt._compat import ipaddress
from salt.utils.dns import (
    _data2rec,
    _data2rec_group,
    _lookup_dig,
    _lookup_drill,
    _lookup_gai,
    _lookup_host,
    _lookup_nslookup,
    _to_port,
    _tree,
    _weighted_order,
    lookup,
)
from salt.utils.odict import OrderedDict
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class DNShelpersCase(TestCase):
    """
    Tests for the parser helpers
    """

    def test_port(self):
        for right in (1, 42, "123", 65535):
            self.assertEqual(_to_port(right), int(right))

        for wrong in (0, 65536, 100000, "not-a-port"):
            self.assertRaises(ValueError, _to_port, wrong)

    def test_tree(self):
        test_map = (
            "ex1.nl",
            "o.1.example.eu",
            "a1a.b2b.c3c.example.com",
            "c3c.example.co.uk",
            "c3c.example.mil.ng",
        )

        res_map = (
            ["ex1.nl"],
            ["o.1.example.eu", "1.example.eu", "example.eu"],
            [
                "a1a.b2b.c3c.example.com",
                "b2b.c3c.example.com",
                "c3c.example.com",
                "example.com",
            ],
            ["c3c.example.co.uk", "example.co.uk"],
            ["c3c.example.mil.ng", "example.mil.ng"],
        )

        for domain, result in zip(test_map, res_map):
            self.assertEqual(_tree(domain), result)

    def test_weight(self):
        recs = [
            [],
            [{"weight": 100, "name": "nescio"}],
            [
                {"weight": 100, "name": "nescio1"},
                {"weight": 100, "name": "nescio2"},
                {"weight": 100, "name": "nescio3"},
                {"weight": 100, "name": "nescio4"},
                {"weight": 100, "name": "nescio5"},
                {"weight": 100, "name": "nescio6"},
                {"weight": 100, "name": "nescio7"},
                {"weight": 100, "name": "nescio8"},
            ],
        ]

        # What are the odds of this tripping over a build
        # 1/(8!^4) builds?
        self.assertNotEqual(
            _weighted_order(list(recs[-1])),
            _weighted_order(list(recs[-1])),
            _weighted_order(list(recs[-1])),
        )

        for recset in recs:
            rs_res = _weighted_order(list(recset))
            self.assertTrue(all(rec["name"] in rs_res for rec in recset))

    def test_data2rec(self):
        right = [
            "10.0.0.1",
            "10 mbox.example.com",
            "10 20 30 example.com",
        ]
        schemas = [
            OrderedDict((("address", ipaddress.IPv4Address),)),
            OrderedDict(
                (
                    ("preference", int),
                    ("name", str),
                )
            ),
            OrderedDict(
                (
                    ("prio", int),
                    ("weight", int),
                    ("port", _to_port),
                    ("name", str),
                )
            ),
        ]

        results = [
            ipaddress.IPv4Address(right[0]),
            {"preference": 10, "name": "mbox.example.com"},
            {"prio": 10, "weight": 20, "port": 30, "name": "example.com"},
        ]

        for rdata, rschema, res in zip(right, schemas, results):
            self.assertEqual(_data2rec(rschema, rdata), res)

        wrong = [
            "not-an-ip",
            "hundred 20 30 interror.example.com",
            "10 toolittle.example.com",
        ]

        for rdata, rschema in zip(wrong, schemas):
            self.assertRaises(ValueError, _data2rec, rschema, rdata)

    def test_data2group(self):
        right = [
            ["10 mbox.example.com"],
            [
                "10 mbox1.example.com",
                "20 mbox2.example.com",
                "20 mbox3.example.com",
                "30 mbox4.example.com",
                "30 mbox5.example.com",
                "30 mbox6.example.com",
            ],
        ]
        rschema = OrderedDict(
            (
                ("prio", int),
                ("srvr", str),
            )
        )

        results = [
            OrderedDict([(10, ["mbox.example.com"])]),
            OrderedDict(
                [
                    (10, ["mbox1.example.com"]),
                    (20, ["mbox2.example.com", "mbox3.example.com"]),
                    (
                        30,
                        ["mbox4.example.com", "mbox5.example.com", "mbox6.example.com"],
                    ),
                ]
            ),
        ]

        for rdata, res in zip(right, results):
            group = _data2rec_group(rschema, rdata, "prio")
            self.assertEqual(group, res)


class DNSlookupsCase(TestCase):
    """
    Test the lookup result parsers

    Note that by far and large the parsers actually
    completely ignore the input name or output content
    only nslookup is bad enough to be an exception to that

    a lookup function
    - raises ValueError when an incorrect DNS type is given
    - returns False upon error
    - returns [*record-data] upon succes/no records

    """

    CMD_RET = {"pid": 12345, "retcode": 0, "stderr": "", "stdout": ""}

    RESULTS = {
        "A": [
            ["10.1.1.1"],  # one-match
            ["10.1.1.1", "10.2.2.2", "10.3.3.3"],  # multi-match
        ],
        "AAAA": [
            ["2a00:a00:b01:c02:d03:e04:f05:111"],  # one-match
            [
                "2a00:a00:b01:c02:d03:e04:f05:111",
                "2a00:a00:b01:c02:d03:e04:f05:222",
                "2a00:a00:b01:c02:d03:e04:f05:333",
            ],  # multi-match
        ],
        "CAA": [['0 issue "exampleca.com"', '0 iodef "mailto:sslabuse@example.com"']],
        "CNAME": [["web.example.com."]],
        "MX": [
            ["10 mx1.example.com."],
            ["10 mx1.example.com.", "20 mx2.example.eu.", "30 mx3.example.nl."],
        ],
        "SSHFP": [
            [
                "1 1 0aabda8af5418108e8a5d3f90f207226b2c89fbe",
                "1 2 500ca871d8e255e01f1261a2370c4e5406b8712f19916d3ab9f86344a67e5597",
                "3 1 a3b605ce6f044617c6077c46a7cd5d17a767f0d5",
                "4 2 0360d0a5a2fa550f972259e7374533add7ac8e5f303322a5b8e208bbc859ab1b",
            ]
        ],
        "TXT": [
            [
                "v=spf1 a include:_spf4.example.com include:mail.example.eu"
                " ip4:10.0.0.0/8 ip6:2a00:a00:b01::/48 ~all"
            ]
        ],
    }

    def _mock_cmd_ret(self, delta_res):
        """
        Take CMD_RET and update it w/(a list of ) delta_res
        Mock cmd.run_all w/it
        :param delta_res: list or dict
        :return: patched cmd.run_all
        """
        if isinstance(delta_res, (list, tuple)):
            test_res = []
            for dres in delta_res:
                tres = self.CMD_RET.copy()
                tres.update(dres)
                test_res.append(tres)

            cmd_mock = MagicMock(side_effect=test_res)
        else:
            test_res = self.CMD_RET.copy()
            test_res.update(delta_res)
            cmd_mock = MagicMock(return_value=test_res)
        return patch.dict(
            salt.utils.dns.__salt__, {"cmd.run_all": cmd_mock}, clear=True
        )

    def _test_cmd_lookup(
        self, lookup_cb, wrong_type, wrong, right, empty=None, secure=None
    ):
        """
        Perform a given battery of tests against a given lookup utilizing cmd.run_all
        :param wrong_type: delta cmd.run_all output for an incorrect DNS type
        :param wrong: delta cmd.run_all output for any error
        :param right: delta cmd.run_all output for outputs in RESULTS
        :param empty: delta cmd.run_all output for anything that won't return matches
        :param secure: delta cmd.run_all output for secured RESULTS
        """
        # wrong
        for _wrong in wrong:
            with self._mock_cmd_ret(_wrong):
                self.assertEqual(lookup_cb("mockq", "A"), False)

        # empty response
        if empty is None:
            empty = {}
        with self._mock_cmd_ret(empty):
            self.assertEqual(lookup_cb("mockq", "AAAA"), [])

        # wrong types
        with self._mock_cmd_ret(wrong_type):
            self.assertRaises(ValueError, lookup_cb, "mockq", "WRONG")

        # Regular outputs
        for rec_t, tests in right.items():
            with self._mock_cmd_ret([dict([("stdout", dres)]) for dres in tests]):
                for test_res in self.RESULTS[rec_t]:
                    if rec_t in ("A", "AAAA", "CNAME", "SSHFP"):
                        rec = "mocksrvr.example.com"
                    else:
                        rec = "example.com"

                    lookup_res = lookup_cb(rec, rec_t)
                    if rec_t == "SSHFP":
                        # Some resolvers 'split' the output and/or capitalize differently.
                        # So we need to workaround that here as well
                        lookup_res = [
                            res[:4] + res[4:].replace(" ", "").lower()
                            for res in lookup_res
                        ]

                    self.assertEqual(
                        lookup_res,
                        test_res,
                        # msg='Error parsing {0} returns'.format(rec_t)
                    )

        if not secure:
            return

        # Regular outputs are insecure outputs (e.g. False)
        for rec_t, tests in right.items():
            with self._mock_cmd_ret([dict([("stdout", dres)]) for dres in tests]):
                for _ in self.RESULTS[rec_t]:
                    self.assertEqual(
                        lookup_cb("mocksrvr.example.com", rec_t, secure=True),
                        False,
                        msg=f"Insecure {rec_t} returns should not be returned",
                    )

        for rec_t, tests in secure.items():
            with self._mock_cmd_ret([dict([("stdout", dres)]) for dres in tests]):
                for test_res in self.RESULTS[rec_t]:
                    self.assertEqual(
                        lookup_cb("mocksrvr.example.com", rec_t, secure=True),
                        test_res,
                        msg=f"Error parsing DNSSEC'd {rec_t} returns",
                    )

    @pytest.mark.skipif(
        not salt.utils.dns.HAS_NSLOOKUP, reason="nslookup is not available"
    )
    @pytest.mark.requires_network
    def test_lookup_with_servers(self):
        rights = {
            "A": [
                "Name:\tmocksrvr.example.com\nAddress: 10.1.1.1",
                "Name:\tmocksrvr.example.com\nAddress: 10.1.1.1\n"
                "Name:\tweb.example.com\nAddress: 10.2.2.2\n"
                "Name:\tweb.example.com\nAddress: 10.3.3.3",
            ],
            "AAAA": [
                "mocksrvr.example.com\thas AAAA address"
                " 2a00:a00:b01:c02:d03:e04:f05:111",
                "mocksrvr.example.com\tcanonical name = web.example.com.\n"
                "web.example.com\thas AAAA address 2a00:a00:b01:c02:d03:e04:f05:111\n"
                "web.example.com\thas AAAA address 2a00:a00:b01:c02:d03:e04:f05:222\n"
                "web.example.com\thas AAAA address 2a00:a00:b01:c02:d03:e04:f05:333",
            ],
            "CNAME": ["mocksrvr.example.com\tcanonical name = web.example.com."],
            "MX": [
                "example.com\tmail exchanger = 10 mx1.example.com.",
                "example.com\tmail exchanger = 10 mx1.example.com.\n"
                "example.com\tmail exchanger = 20 mx2.example.eu.\n"
                "example.com\tmail exchanger = 30 mx3.example.nl.",
            ],
            "TXT": [
                'example.com\ttext = "v=spf1 a include:_spf4.example.com'
                ' include:mail.example.eu ip4:10.0.0.0/8 ip6:2a00:a00:b01::/48 ~all"'
            ],
        }

        for rec_t, tests in rights.items():
            with self._mock_cmd_ret([dict([("stdout", dres)]) for dres in tests]):
                for test_res in self.RESULTS[rec_t]:
                    if rec_t in ("A", "AAAA", "CNAME"):
                        rec = "mocksrvr.example.com"
                    else:
                        rec = "example.com"
                    self.assertEqual(
                        lookup(rec, rec_t, method="nslookup", servers="8.8.8.8"),
                        test_res,
                    )

    @pytest.mark.skipif(not salt.utils.dns.HAS_DIG, reason="dig is not available")
    def test_dig_options(self):
        cmd = f"dig {salt.utils.dns.DIG_OPTIONS} -v"
        cmd = salt.modules.cmdmod.retcode(
            cmd, python_shell=False, output_loglevel="quiet"
        )
        self.assertEqual(cmd, 0)

    def test_dig(self):
        wrong_type = {"retcode": 0, "stderr": ";; Warning, ignoring invalid type ABC"}

        wrongs = [
            {
                "retcode": 9,
                "stderr": ";; connection timed out; no servers could be reached",
            },
        ]

        # example returns for dig +search +fail +noall +answer +noclass +nosplit +nottl -t {rtype} {name}
        rights = {
            "A": [
                "mocksrvr.example.com.\tA\t10.1.1.1",
                "web.example.com.\t\tA\t10.1.1.1\n"
                "web.example.com.\t\tA\t10.2.2.2\n"
                "web.example.com.\t\tA\t10.3.3.3",
            ],
            "AAAA": [
                "mocksrvr.example.com.\tA\t2a00:a00:b01:c02:d03:e04:f05:111",
                "mocksrvr.example.com.\tCNAME\tweb.example.com.\n"
                "web.example.com.\t\tAAAA\t2a00:a00:b01:c02:d03:e04:f05:111\n"
                "web.example.com.\t\tAAAA\t2a00:a00:b01:c02:d03:e04:f05:222\n"
                "web.example.com.\t\tAAAA\t2a00:a00:b01:c02:d03:e04:f05:333",
            ],
            "CAA": [
                'example.com.\t\tCAA\t0 issue "exampleca.com"\n'
                'example.com.\t\tCAA\t0 iodef "mailto:sslabuse@example.com"'
            ],
            "CNAME": ["mocksrvr.example.com.\tCNAME\tweb.example.com."],
            "MX": [
                "example.com.\t\tMX\t10 mx1.example.com.",
                "example.com.\t\tMX\t10 mx1.example.com.\nexample.com.\t\tMX\t20"
                " mx2.example.eu.\nexample.com.\t\tMX\t30 mx3.example.nl.",
            ],
            "SSHFP": [
                "mocksrvr.example.com.\tSSHFP\t1 1"
                " 0AABDA8AF5418108E8A5D3F90F207226B2C89FBE\nmocksrvr.example.com.\tSSHFP\t1"
                " 2 500CA871D8E255E01F1261A2370C4E5406B8712F19916D3AB9F86344A67E5597\nmocksrvr.example.com.\tSSHFP\t3"
                " 1 A3B605CE6F044617C6077C46A7CD5D17A767F0D5\nmocksrvr.example.com.\tSSHFP\t4"
                " 2 0360D0A5A2FA550F972259E7374533ADD7AC8E5F303322A5B8E208BBC859AB1B"
            ],
            "TXT": [
                'example.com.\tTXT\t"v=spf1 a include:_spf4.example.com'
                ' include:mail.example.eu ip4:10.0.0.0/8 ip6:2a00:a00:b01::/48 ~all"'
            ],
        }

        secure = {
            "A": [
                "mocksrvr.example.com.\tA\t10.1.1.1\nmocksrvr.example.com.\tRRSIG\tA 8"
                " 3 7200 20170420000000 20170330000000 1629 example.com."
                " Hv4p37EF55LKBxUNYpnhWiEYqfmMct0z0WgDJyG5reqYfl+z4HX/kaoi"
                " Wr2iCYuYeB4Le7BgnMSb77UGHPWE7lCQ8z5gkgJ9rCDrooJzSTVdnHfw"
                " 1JQ7txRSp8Rj2GLf/L3Ytuo6nNZTV7bWUkfhOs61DAcOPHYZiX8rVhIh UAE=",
                "web.example.com.\t\tA\t10.1.1.1\nweb.example.com.\t\tA\t10.2.2.2\nweb.example.com.\t\tA\t10.3.3.3\nweb.example.com.\tRRSIG\tA"
                " 8 3 7200 20170420000000 20170330000000 1629 example.com."
                " Hv4p37EF55LKBxUNYpnhWiEYqfmMct0z0WgDJyG5reqYfl+z4HX/kaoi"
                " Wr2iCYuYeB4Le7BgnMSb77UGHPWE7lCQ8z5gkgJ9rCDrooJzSTVdnHfw"
                " 1JQ7txRSp8Rj2GLf/L3Ytuo6nNZTV7bWUkfhOs61DAcOPHYZiX8rVhIh UAE=",
            ]
        }

        self._test_cmd_lookup(
            _lookup_dig,
            wrong=wrongs,
            right=rights,
            wrong_type=wrong_type,
            secure=secure,
        )

    def test_drill(self):
        # all Drill returns look like this
        RES_TMPL = textwrap.dedent(
            """\
            ;; ->>HEADER<<- opcode: QUERY, rcode: NOERROR, id: 58233
            ;; flags: qr rd ra ; QUERY: 1, ANSWER: 1, AUTHORITY: 0, ADDITIONAL: 0
            ;; QUESTION SECTION:
            ;; mocksrvr.example.com.\tIN\tA

            ;; ANSWER SECTION:
            {}

            ;; AUTHORITY SECTION:

            ;; ADDITIONAL SECTION:

            ;; Query time: 37 msec
            ;; SERVER: 10.100.150.129
            ;; WHEN: Tue Apr  4 19:03:51 2017
            ;; MSG SIZE  rcvd: 50
        """
        )

        # Not even a different retcode!?
        wrong_type = {
            "stdout": RES_TMPL.format("mocksrvr.example.com.\t4404\tIN\tA\t10.1.1.1\n")
        }

        wrongs = [
            {
                "retcode": 1,
                "stderr": (
                    "Error: error sending query: No (valid) nameservers defined in the"
                    " resolver"
                ),
            }
        ]

        # example returns for drill {rtype} {name}
        rights = {
            "A": [
                "mocksrvr.example.com.\t4404\tIN\tA\t10.1.1.1\n",
                "web.example.com.\t4404\tIN\tA\t10.1.1.1\n"
                "web.example.com.\t4404\tIN\tA\t10.2.2.2\n"
                "web.example.com.\t4404\tIN\tA\t10.3.3.3",
            ],
            "AAAA": [
                "mocksrvr.example.com.\t4404\tIN\tAAAA\t2a00:a00:b01:c02:d03:e04:f05:111",
                "mocksrvr.example.com.\t4404\tIN\tCNAME\tweb.example.com.\n"
                "web.example.com.\t4404\tIN\tAAAA\t2a00:a00:b01:c02:d03:e04:f05:111\n"
                "web.example.com.\t4404\tIN\tAAAA\t2a00:a00:b01:c02:d03:e04:f05:222\n"
                "web.example.com.\t4404\tIN\tAAAA\t2a00:a00:b01:c02:d03:e04:f05:333",
            ],
            "CAA": [
                'example.com.\t1144\tIN\tCAA\t0 issue "exampleca.com"\n'
                'example.com.\t1144\tIN\tCAA\t0 iodef "mailto:sslabuse@example.com"'
            ],
            "CNAME": ["mocksrvr.example.com.\t4404\tIN\tCNAME\tweb.example.com."],
            "MX": [
                "example.com.\t4404\tIN\tMX\t10 mx1.example.com.",
                "example.com.\t4404\tIN\tMX\t10 mx1.example.com.\n"
                "example.com.\t4404\tIN\tMX\t20 mx2.example.eu.\n"
                "example.com.\t4404\tIN\tMX\t30 mx3.example.nl.",
            ],
            "SSHFP": [
                "mocksrvr.example.com.\t3339\tIN\tSSHFP\t1 1"
                " 0aabda8af5418108e8a5d3f90f207226b2c89fbe\nmocksrvr.example.com.\t3339\tIN\tSSHFP\t1"
                " 2 500ca871d8e255e01f1261a2370c4e5406b8712f19916d3ab9f86344a67e5597\nmocksrvr.example.com.\t3339\tIN\tSSHFP\t3"
                " 1 a3b605ce6f044617c6077c46a7cd5d17a767f0d5\nmocksrvr.example.com.\t3339\tIN\tSSHFP\t4"
                " 2 0360d0a5a2fa550f972259e7374533add7ac8e5f303322a5b8e208bbc859ab1b"
            ],
            "TXT": [
                'example.com.\t4404\tIN\tTXT\t"v=spf1 a include:_spf4.example.com'
                ' include:mail.example.eu ip4:10.0.0.0/8 ip6:2a00:a00:b01::/48 ~all"'
            ],
        }

        secure = {
            "A": [
                "mocksrvr.example.com.\t4404\tIN\tA\t10.1.1.1\nmocksrvr.example.com.\t4404\tIN\tRRSIG\tA"
                " 8 3 7200 20170420000000 20170330000000 1629 example.com."
                " Hv4p37EF55LKBxUNYpnhWiEYqfmMct0z0WgDJyG5reqYfl+z4HX/kaoi"
                " Wr2iCYuYeB4Le7BgnMSb77UGHPWE7lCQ8z5gkgJ9rCDrooJzSTVdnHfw"
                " 1JQ7txRSp8Rj2GLf/L3Ytuo6nNZTV7bWUkfhOs61DAcOPHYZiX8rVhIh UAE=",
                "web.example.com.\t4404\tIN\tA\t10.1.1.1\nweb.example.com.\t4404\tIN\tA\t10.2.2.2\nweb.example.com.\t4404\tIN\tA\t10.3.3.3\nweb.example.com.\t4404\tIN\tRRSIG\tA"
                " 8 3 7200 20170420000000 20170330000000 1629 example.com."
                " Hv4p37EF55LKBxUNYpnhWiEYqfmMct0z0WgDJyG5reqYfl+z4HX/kaoi"
                " Wr2iCYuYeB4Le7BgnMSb77UGHPWE7lCQ8z5gkgJ9rCDrooJzSTVdnHfw"
                " 1JQ7txRSp8Rj2GLf/L3Ytuo6nNZTV7bWUkfhOs61DAcOPHYZiX8rVhIh UAE=",
            ]
        }

        for rec_d in rights, secure:
            for rec_t, tests in rec_d.items():
                for idx, test in enumerate(tests):
                    rec_d[rec_t][idx] = RES_TMPL.format(test)

        self._test_cmd_lookup(
            _lookup_drill,
            wrong_type=wrong_type,
            wrong=wrongs,
            right=rights,
            secure=secure,
        )

    def test_gai(self):
        # wrong type
        self.assertRaises(ValueError, _lookup_gai, "mockq", "WRONG")

        # wrong
        with patch.object(
            socket, "getaddrinfo", MagicMock(side_effect=socket.gaierror)
        ):
            for rec_t in ("A", "AAAA"):
                self.assertEqual(False, _lookup_gai("mockq", rec_t))

        # example returns from getaddrinfo
        right = {
            "A": [
                [(2, 3, 3, "", ("10.1.1.1", 0))],
                [
                    (2, 3, 3, "", ("10.1.1.1", 0)),
                    (2, 3, 3, "", ("10.2.2.2", 0)),
                    (2, 3, 3, "", ("10.3.3.3", 0)),
                ],
            ],
            "AAAA": [
                [(10, 3, 3, "", ("2a00:a00:b01:c02:d03:e04:f05:111", 0, 0, 0))],
                [
                    (10, 3, 3, "", ("2a00:a00:b01:c02:d03:e04:f05:111", 0, 0, 0)),
                    (10, 3, 3, "", ("2a00:a00:b01:c02:d03:e04:f05:222", 0, 0, 0)),
                    (10, 3, 3, "", ("2a00:a00:b01:c02:d03:e04:f05:333", 0, 0, 0)),
                ],
            ],
        }

        for rec_t, tests in right.items():
            with patch.object(socket, "getaddrinfo", MagicMock(side_effect=tests)):
                for test_res in self.RESULTS[rec_t]:
                    self.assertEqual(
                        _lookup_gai("mockq", rec_t),
                        test_res,
                        msg=f"Error parsing {rec_t} returns",
                    )

    def test_host(self):
        wrong_type = {"retcode": 9, "stderr": "host: invalid type: WRONG"}

        wrongs = [
            {
                "retcode": 9,
                "stderr": ";; connection timed out; no servers could be reached",
            }
        ]

        empty = {"stdout": "www.example.com has no MX record"}

        # example returns for host -t {rdtype} {name}
        rights = {
            "A": [
                "mocksrvr.example.com has address 10.1.1.1",
                "web.example.com has address 10.1.1.1\n"
                "web.example.com has address 10.2.2.2\n"
                "web.example.com has address 10.3.3.3",
            ],
            "AAAA": [
                "mocksrvr.example.com has IPv6 address"
                " 2a00:a00:b01:c02:d03:e04:f05:111",
                "mocksrvr.example.com is an alias for web.example.com.\n"
                "web.example.com has IPv6 address 2a00:a00:b01:c02:d03:e04:f05:111\n"
                "web.example.com has IPv6 address 2a00:a00:b01:c02:d03:e04:f05:222\n"
                "web.example.com has IPv6 address 2a00:a00:b01:c02:d03:e04:f05:333",
            ],
            "CAA": [
                'example.com has CAA record 0 issue "exampleca.com"\n'
                'example.com has CAA record 0 iodef "mailto:sslabuse@example.com"'
            ],
            "CNAME": ["mocksrvr.example.com is an alias for web.example.com."],
            "MX": [
                "example.com mail is handled by 10 mx1.example.com.",
                "example.com mail is handled by 10 mx1.example.com.\n"
                "example.com mail is handled by 20 mx2.example.eu.\n"
                "example.com mail is handled by 30 mx3.example.nl.",
            ],
            "SSHFP": [
                "mocksrvr.example.com has SSHFP record 1 1"
                " 0AABDA8AF5418108E8A5D3F90F207226B2C89FBE\nmocksrvr.example.com has"
                " SSHFP record 1 2"
                " 500CA871D8E255E01F1261A2370C4E5406B8712F19916D3AB9F86344"
                " A67E5597\nmocksrvr.example.com has SSHFP record 3 1"
                " A3B605CE6F044617C6077C46A7CD5D17A767F0D5\nmocksrvr.example.com has"
                " SSHFP record 4 2"
                " 0360D0A5A2FA550F972259E7374533ADD7AC8E5F303322A5B8E208BB C859AB1B"
            ],
            "TXT": [
                'example.com descriptive text "v=spf1 a include:_spf4.example.com'
                ' include:mail.example.eu ip4:10.0.0.0/8 ip6:2a00:a00:b01::/48 ~all"'
            ],
        }

        self._test_cmd_lookup(
            _lookup_host, wrong_type=wrong_type, wrong=wrongs, right=rights, empty=empty
        )

    def test_nslookup(self):
        # all nslookup returns look like this
        RES_TMPL = textwrap.dedent(
            """\
            Server:\t\t10.11.12.13
            Address:\t10.11.12.13#53

            Non-authoritative answer:
            {}

            Authoritative answers can be found from:
        """
        )

        wrong_type = {
            "stdout": "unknown query type: WRONG"
            + RES_TMPL.format("Name:\tmocksrvr.example.com\nAddress: 10.1.1.1")
        }

        wrongs = [
            {
                "retcode": 1,
                "stdout": ";; connection timed out; no servers could be reached",
            }
        ]

        empty = {"stdout": RES_TMPL.format("*** Can't find www.google.com: No answer")}

        # Example returns of nslookup -query={rdype} {name}
        rights = {
            "A": [
                "Name:\tmocksrvr.example.com\nAddress: 10.1.1.1",
                "Name:\tmocksrvr.example.com\nAddress: 10.1.1.1\n"
                "Name:\tweb.example.com\nAddress: 10.2.2.2\n"
                "Name:\tweb.example.com\nAddress: 10.3.3.3",
            ],
            "AAAA": [
                "mocksrvr.example.com\thas AAAA address"
                " 2a00:a00:b01:c02:d03:e04:f05:111",
                "mocksrvr.example.com\tcanonical name = web.example.com.\n"
                "web.example.com\thas AAAA address 2a00:a00:b01:c02:d03:e04:f05:111\n"
                "web.example.com\thas AAAA address 2a00:a00:b01:c02:d03:e04:f05:222\n"
                "web.example.com\thas AAAA address 2a00:a00:b01:c02:d03:e04:f05:333",
            ],
            "CAA": [
                'example.com\trdata_257 = 0 issue "exampleca.com"\n'
                'example.com\trdata_257 = 0 iodef "mailto:sslabuse@example.com"'
            ],
            "CNAME": ["mocksrvr.example.com\tcanonical name = web.example.com."],
            "MX": [
                "example.com\tmail exchanger = 10 mx1.example.com.",
                "example.com\tmail exchanger = 10 mx1.example.com.\n"
                "example.com\tmail exchanger = 20 mx2.example.eu.\n"
                "example.com\tmail exchanger = 30 mx3.example.nl.",
            ],
            "SSHFP": [
                "mocksrvr.example.com\trdata_44 = 1 1"
                " 0AABDA8AF5418108E8A5D3F90F207226B2C89FBE\nmocksrvr.example.com\trdata_44"
                " = 1 2 500CA871D8E255E01F1261A2370C4E5406B8712F19916D3AB9F86344"
                " A67E5597\nmocksrvr.example.com\trdata_44 = 3 1"
                " A3B605CE6F044617C6077C46A7CD5D17A767F0D5\nmocksrvr.example.com\trdata_44"
                " = 4 2 0360D0A5A2FA550F972259E7374533ADD7AC8E5F303322A5B8E208BB"
                " C859AB1B"
            ],
            "TXT": [
                'example.com\ttext = "v=spf1 a include:_spf4.example.com'
                ' include:mail.example.eu ip4:10.0.0.0/8 ip6:2a00:a00:b01::/48 ~all"'
            ],
        }

        for rec_t, tests in rights.items():
            for idx, test in enumerate(tests):
                rights[rec_t][idx] = RES_TMPL.format(test)

        self._test_cmd_lookup(
            _lookup_nslookup,
            wrong_type=wrong_type,
            wrong=wrongs,
            right=rights,
            empty=empty,
        )



"""
Apache state

.. versionadded:: 2014.7.0

Allows for inputting a yaml dictionary into a file for apache configuration
files.

The variable ``this`` is special and signifies what should be included with
the above word between angle brackets (<>).

.. code-block:: yaml

    /etc/httpd/conf.d/website.com.conf:
      apache.configfile:
        - config:
          - VirtualHost:
              this: '*:80'
              ServerName:
                - website.com
              ServerAlias:
                - www.website.com
                - dev.website.com
              ErrorLog: logs/website.com-error_log
              CustomLog: logs/website.com-access_log combined
              DocumentRoot: /var/www/vhosts/website.com
              Directory:
                this: /var/www/vhosts/website.com
                Order: Deny,Allow
                Deny from: all
                Allow from:
                  - 127.0.0.1
                  - 192.168.100.0/24
                Options:
                  - Indexes
                  - FollowSymlinks
                AllowOverride: All

.. versionchanged:: 2018.3.0

Allows having the same section container multiple times (e.g. <Directory /path/to/dir>).

YAML structure stays the same only replace dictionary with a list.

When a section container does not have mandatory attribute, such as <Else>,
it still needs keyword ``this`` with empty string (or "\b" if nicer output is required - without space).

.. code-block:: yaml

    /etc/httpd/conf.d/website.com.conf:
      apache.configfile:
        - config:
          - VirtualHost:
              - this: '*:80'
              - ServerName:
                - website.com
              - DocumentRoot: /var/www/vhosts/website.com
              - Directory:
                  this: /var/www/vhosts/website.com
                  Order: Deny,Allow
                  Deny from: all
                  Allow from:
                    - 127.0.0.1
                    - 192.168.100.0/24
                  Options:
                    - Indexes
                    - FollowSymlinks
                  AllowOverride: All
              - Directory:
                - this: /var/www/vhosts/website.com/private
                - Order: Deny,Allow
                - Deny from: all
                - Allow from:
                  - 127.0.0.1
                  - 192.168.100.0/24
                - If:
                    this: some condition
                    do: something
                - Else:
                    this:
                    do: something else
                - Else:
                    this: "\b"
                    do: another thing
"""

import os

import salt.utils.files
import salt.utils.stringutils

__deprecated__ = (
    3009,
    "apache",
    "https://github.com/salt-extensions/saltext-apache",
)


def __virtual__():
    if "apache.config" in __salt__:
        return True
    return (False, "apache module could not be loaded")


def configfile(name, config):
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    configs = __salt__["apache.config"](name, config, edit=False)
    current_configs = ""
    if os.path.exists(name):
        with salt.utils.files.fopen(name) as config_file:
            current_configs = salt.utils.stringutils.to_unicode(config_file.read())

    if configs.strip() == current_configs.strip():
        ret["result"] = True
        ret["comment"] = "Configuration is up to date."
        return ret
    elif __opts__["test"]:
        ret["comment"] = "Configuration will update."
        ret["changes"] = {"old": current_configs, "new": configs}
        ret["result"] = None
        return ret

    try:
        with salt.utils.files.fopen(name, "w") as config_file:
            print(salt.utils.stringutils.to_str(configs), file=config_file)
        ret["changes"] = {"old": current_configs, "new": configs}
        ret["result"] = True
        ret["comment"] = "Successfully created configuration."
    except Exception as exc:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = "Failed to create apache configuration."

    return ret
